# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Implementations of unittest features from the future."""

import unittest


def unittest_has(method):
    """Does `unittest.TestCase` have `method` defined?"""
    return hasattr(unittest.TestCase, method)


class TestCase(unittest.TestCase):
    """Just like unittest.TestCase, but with assert methods added.

    Designed to be compatible with 3.1 unittest.  Methods are only defined if
    `unittest` doesn't have them.

    """
    # pylint: disable=signature-differs, deprecated-method

    if not unittest_has('assertCountEqual'):
        def assertCountEqual(self, *args, **kwargs):
            return self.assertItemsEqual(*args, **kwargs)

    if not unittest_has('assertRaisesRegex'):
        def assertRaisesRegex(self, *args, **kwargs):
            return self.assertRaisesRegexp(*args, **kwargs)

    if not unittest_has('assertRegex'):
        def assertRegex(self, *args, **kwargs):
            return self.assertRegexpMatches(*args, **kwargs)



from django.conf import settings
from django.core.cache import caches

try:
    from django.utils.module_loading import import_string
except ImportError:
    try:
        from django.utils.module_loading import import_by_path as import_string
    except ImportError:
        from django.utils.importlib import import_module as import_string


class CacheRouter:
    def get_cache(self, **kwargs):
        return caches[self.get_cache_name(**kwargs)]

    def get_cache_name(self, **kwargs):
        return "default"


def _get_router():
    if getattr(settings, "CACHE_ROUTER", None):
        router = import_string(settings.CACHE_ROUTER)
        return router()
    else:
        return CacheRouter()


router = _get_router()



import unittest
from django.db import models as dbmodels

from scarlet.versioning.models import VersionView

from . import models


class ModelStructureTests(unittest.TestCase):
    def testCopyAttrs(self):
        O1 = type(
            "C1", (VersionView,), {"__module__": __name__, "new_attr": "something"}
        )
        o = O1()
        self.assertFalse(hasattr(o._meta._version_model, "new_attr"))

        O2 = type(
            "C2",
            (VersionView,),
            {
                "__module__": __name__,
                "new_attr": "something",
                "_copy_extra_attrs": ["new_attr"],
            },
        )
        o2 = O2()
        self.assertTrue(hasattr(o2._meta._version_model, "new_attr"))
        self.assertEqual(o2._meta._version_model.new_attr, "something")

    def testCustomBadBaseModel(self):
        """
        _base_model should be a subclass of BaseModel
        """

        with self.assertRaises(AssertionError):

            class BadCustomModel(dbmodels.Model):
                reg_number = dbmodels.CharField(max_length=20)

            class BadGun(VersionView):
                name = dbmodels.CharField(max_length=20)

                class Meta:
                    _base_model = BadCustomModel

    def testNoConcreteModels(self):
        """
        Version Views can't inherit from concrete models
        """

        with self.assertRaises(TypeError):

            class Bad(VersionView, models.ConcreteModel):
                test = dbmodels.CharField(max_length=20)

    def testTooManyBaseVersionedModels(self):
        class V1(models.BaseVersionedModel):
            name = dbmodels.CharField(max_length=255)

            class Meta:
                abstract = True

        class V2(models.BaseVersionedModel):
            code = dbmodels.CharField(max_length=255)

            class Meta:
                abstract = True

        with self.assertRaises(TypeError):

            class BadBases(VersionView, V1, V2):
                test = dbmodels.CharField(max_length=20)

    def testM2MOnVersioned(self):
        with self.assertRaises(TypeError):

            class BadM2M(VersionView):
                m2 = dbmodels.ManyToManyField("Author")

    def testNoBaseModel(self):
        with self.assertRaises(TypeError):

            class BadBase(VersionView, models.BaseModel):
                name = dbmodels.CharField(max_length=255)



"""All pytest-django fixtures"""

from __future__ import with_statement

import os
import warnings
from contextlib import contextmanager
from functools import partial

import pytest

from . import live_server_helper
from .django_compat import is_django_unittest
from .lazy_django import skip_if_no_django

__all__ = [
    "django_db_setup",
    "db",
    "transactional_db",
    "django_db_reset_sequences",
    "admin_user",
    "django_user_model",
    "django_username_field",
    "client",
    "admin_client",
    "rf",
    "settings",
    "live_server",
    "_live_server_helper",
    "django_assert_num_queries",
    "django_assert_max_num_queries",
]


@pytest.fixture(scope="session")
def django_db_modify_db_settings_tox_suffix():
    skip_if_no_django()

    tox_environment = os.getenv("TOX_PARALLEL_ENV")
    if tox_environment:
        # Put a suffix like _py27-django21 on tox workers
        _set_suffix_to_test_databases(suffix=tox_environment)


@pytest.fixture(scope="session")
def django_db_modify_db_settings_xdist_suffix(request):
    skip_if_no_django()

    xdist_suffix = getattr(request.config, "slaveinput", {}).get("slaveid")
    if xdist_suffix:
        # Put a suffix like _gw0, _gw1 etc on xdist processes
        _set_suffix_to_test_databases(suffix=xdist_suffix)


@pytest.fixture(scope="session")
def django_db_modify_db_settings_parallel_suffix(
    django_db_modify_db_settings_tox_suffix,
    django_db_modify_db_settings_xdist_suffix,
):
    skip_if_no_django()


@pytest.fixture(scope="session")
def django_db_modify_db_settings(django_db_modify_db_settings_parallel_suffix):
    skip_if_no_django()


@pytest.fixture(scope="session")
def django_db_use_migrations(request):
    return not request.config.getvalue("nomigrations")


@pytest.fixture(scope="session")
def django_db_keepdb(request):
    return request.config.getvalue("reuse_db")


@pytest.fixture(scope="session")
def django_db_createdb(request):
    return request.config.getvalue("create_db")


@pytest.fixture(scope="session")
def django_db_setup(
    request,
    django_test_environment,
    django_db_blocker,
    django_db_use_migrations,
    django_db_keepdb,
    django_db_createdb,
    django_db_modify_db_settings,
):
    """Top level fixture to ensure test databases are available"""
    from .compat import setup_databases, teardown_databases

    setup_databases_args = {}

    if not django_db_use_migrations:
        _disable_native_migrations()

    if django_db_keepdb and not django_db_createdb:
        setup_databases_args["keepdb"] = True

    with django_db_blocker.unblock():
        db_cfg = setup_databases(
            verbosity=request.config.option.verbose,
            interactive=False,
            **setup_databases_args
        )

    def teardown_database():
        with django_db_blocker.unblock():
            try:
                teardown_databases(db_cfg, verbosity=request.config.option.verbose)
            except Exception as exc:
                request.node.warn(
                    pytest.PytestWarning(
                        "Error when trying to teardown test databases: %r" % exc
                    )
                )

    if not django_db_keepdb:
        request.addfinalizer(teardown_database)


def _django_db_fixture_helper(
    request, django_db_blocker, transactional=False, reset_sequences=False
):
    if is_django_unittest(request):
        return

    if not transactional and "live_server" in request.fixturenames:
        # Do nothing, we get called with transactional=True, too.
        return

    django_db_blocker.unblock()
    request.addfinalizer(django_db_blocker.restore)

    if transactional:
        from django.test import TransactionTestCase as django_case

        if reset_sequences:

            class ResetSequenceTestCase(django_case):
                reset_sequences = True

            django_case = ResetSequenceTestCase
    else:
        from django.test import TestCase as django_case

    test_case = django_case(methodName="__init__")
    test_case._pre_setup()
    request.addfinalizer(test_case._post_teardown)


def _disable_native_migrations():
    from django.conf import settings
    from django.core.management.commands import migrate

    from .migrations import DisableMigrations

    settings.MIGRATION_MODULES = DisableMigrations()

    class MigrateSilentCommand(migrate.Command):
        def handle(self, *args, **kwargs):
            kwargs["verbosity"] = 0
            return super(MigrateSilentCommand, self).handle(*args, **kwargs)

    migrate.Command = MigrateSilentCommand


def _set_suffix_to_test_databases(suffix):
    from django.conf import settings

    for db_settings in settings.DATABASES.values():
        test_name = db_settings.get("TEST", {}).get("NAME")

        if not test_name:
            if db_settings["ENGINE"] == "django.db.backends.sqlite3":
                continue
            test_name = "test_{}".format(db_settings["NAME"])

        if test_name == ":memory:":
            continue

        db_settings.setdefault("TEST", {})
        db_settings["TEST"]["NAME"] = "{}_{}".format(test_name, suffix)


# ############### User visible fixtures ################


@pytest.fixture(scope="function")
def db(request, django_db_setup, django_db_blocker):
    """Require a django test database.

    This database will be setup with the default fixtures and will have
    the transaction management disabled. At the end of the test the outer
    transaction that wraps the test itself will be rolled back to undo any
    changes to the database (in case the backend supports transactions).
    This is more limited than the ``transactional_db`` resource but
    faster.

    If multiple database fixtures are requested, they take precedence
    over each other in the following order (the last one wins): ``db``,
    ``transactional_db``, ``django_db_reset_sequences``.
    """
    if "django_db_reset_sequences" in request.fixturenames:
        request.getfixturevalue("django_db_reset_sequences")
    if (
        "transactional_db" in request.fixturenames
        or "live_server" in request.fixturenames
    ):
        request.getfixturevalue("transactional_db")
    else:
        _django_db_fixture_helper(request, django_db_blocker, transactional=False)


@pytest.fixture(scope="function")
def transactional_db(request, django_db_setup, django_db_blocker):
    """Require a django test database with transaction support.

    This will re-initialise the django database for each test and is
    thus slower than the normal ``db`` fixture.

    If you want to use the database with transactions you must request
    this resource.

    If multiple database fixtures are requested, they take precedence
    over each other in the following order (the last one wins): ``db``,
    ``transactional_db``, ``django_db_reset_sequences``.
    """
    if "django_db_reset_sequences" in request.fixturenames:
        request.getfixturevalue("django_db_reset_sequences")
    _django_db_fixture_helper(request, django_db_blocker, transactional=True)


@pytest.fixture(scope="function")
def django_db_reset_sequences(request, django_db_setup, django_db_blocker):
    """Require a transactional test database with sequence reset support.

    This behaves like the ``transactional_db`` fixture, with the addition
    of enforcing a reset of all auto increment sequences.  If the enquiring
    test relies on such values (e.g. ids as primary keys), you should
    request this resource to ensure they are consistent across tests.

    If multiple database fixtures are requested, they take precedence
    over each other in the following order (the last one wins): ``db``,
    ``transactional_db``, ``django_db_reset_sequences``.
    """
    _django_db_fixture_helper(
        request, django_db_blocker, transactional=True, reset_sequences=True
    )


@pytest.fixture()
def client():
    """A Django test client instance."""
    skip_if_no_django()

    from django.test.client import Client

    return Client()


@pytest.fixture()
def django_user_model(db):
    """The class of Django's user model."""
    from django.contrib.auth import get_user_model

    return get_user_model()


@pytest.fixture()
def django_username_field(django_user_model):
    """The fieldname for the username used with Django's user model."""
    return django_user_model.USERNAME_FIELD


@pytest.fixture()
def admin_user(db, django_user_model, django_username_field):
    """A Django admin user.

    This uses an existing user with username "admin", or creates a new one with
    password "password".
    """
    UserModel = django_user_model
    username_field = django_username_field
    username = "admin@example.com" if username_field == "email" else "admin"

    try:
        user = UserModel._default_manager.get(**{username_field: username})
    except UserModel.DoesNotExist:
        extra_fields = {}
        if username_field not in ("username", "email"):
            extra_fields[username_field] = "admin"
        user = UserModel._default_manager.create_superuser(
            username, "admin@example.com", "password", **extra_fields
        )
    return user


@pytest.fixture()
def admin_client(db, admin_user):
    """A Django test client logged in as an admin user."""
    from django.test.client import Client

    client = Client()
    client.login(username=admin_user.username, password="password")
    return client


@pytest.fixture()
def rf():
    """RequestFactory instance"""
    skip_if_no_django()

    from django.test.client import RequestFactory

    return RequestFactory()


class SettingsWrapper(object):
    _to_restore = []

    def __delattr__(self, attr):
        from django.test import override_settings

        override = override_settings()
        override.enable()
        from django.conf import settings

        delattr(settings, attr)

        self._to_restore.append(override)

    def __setattr__(self, attr, value):
        from django.test import override_settings

        override = override_settings(**{attr: value})
        override.enable()
        self._to_restore.append(override)

    def __getattr__(self, item):
        from django.conf import settings

        return getattr(settings, item)

    def finalize(self):
        for override in reversed(self._to_restore):
            override.disable()

        del self._to_restore[:]


@pytest.yield_fixture()
def settings():
    """A Django settings object which restores changes after the testrun"""
    skip_if_no_django()

    wrapper = SettingsWrapper()
    yield wrapper
    wrapper.finalize()


@pytest.fixture(scope="session")
def live_server(request):
    """Run a live Django server in the background during tests

    The address the server is started from is taken from the
    --liveserver command line option or if this is not provided from
    the DJANGO_LIVE_TEST_SERVER_ADDRESS environment variable.  If
    neither is provided ``localhost:8081,8100-8200`` is used.  See the
    Django documentation for its full syntax.

    NOTE: If the live server needs database access to handle a request
          your test will have to request database access.  Furthermore
          when the tests want to see data added by the live-server (or
          the other way around) transactional database access will be
          needed as data inside a transaction is not shared between
          the live server and test code.

          Static assets will be automatically served when
          ``django.contrib.staticfiles`` is available in INSTALLED_APPS.
    """
    skip_if_no_django()

    import django

    addr = request.config.getvalue("liveserver") or os.getenv(
        "DJANGO_LIVE_TEST_SERVER_ADDRESS"
    )

    if addr and ":" in addr:
        if django.VERSION >= (1, 11):
            ports = addr.split(":")[1]
            if "-" in ports or "," in ports:
                warnings.warn(
                    "Specifying multiple live server ports is not supported "
                    "in Django 1.11. This will be an error in a future "
                    "pytest-django release."
                )

    if not addr:
        if django.VERSION < (1, 11):
            addr = "localhost:8081,8100-8200"
        else:
            addr = "localhost"

    server = live_server_helper.LiveServer(addr)
    request.addfinalizer(server.stop)
    return server


@pytest.fixture(autouse=True, scope="function")
def _live_server_helper(request):
    """Helper to make live_server work, internal to pytest-django.

    This helper will dynamically request the transactional_db fixture
    for a test which uses the live_server fixture.  This allows the
    server and test to access the database without having to mark
    this explicitly which is handy since it is usually required and
    matches the Django behaviour.

    The separate helper is required since live_server can not request
    transactional_db directly since it is session scoped instead of
    function-scoped.

    It will also override settings only for the duration of the test.
    """
    if "live_server" not in request.fixturenames:
        return

    request.getfixturevalue("transactional_db")

    live_server = request.getfixturevalue("live_server")
    live_server._live_server_modified_settings.enable()
    request.addfinalizer(live_server._live_server_modified_settings.disable)


@contextmanager
def _assert_num_queries(config, num, exact=True, connection=None, info=None):
    from django.test.utils import CaptureQueriesContext

    if connection is None:
        from django.db import connection

    verbose = config.getoption("verbose") > 0
    with CaptureQueriesContext(connection) as context:
        yield context
        num_performed = len(context)
        if exact:
            failed = num != num_performed
        else:
            failed = num_performed > num
        if failed:
            msg = "Expected to perform {} queries {}{}".format(
                num,
                "" if exact else "or less ",
                "but {} done".format(
                    num_performed == 1 and "1 was" or "%d were" % (num_performed,)
                ),
            )
            if info:
                msg += "\n{}".format(info)
            if verbose:
                sqls = (q["sql"] for q in context.captured_queries)
                msg += "\n\nQueries:\n========\n\n%s" % "\n\n".join(sqls)
            else:
                msg += " (add -v option to show queries)"
            pytest.fail(msg)


@pytest.fixture(scope="function")
def django_assert_num_queries(pytestconfig):
    return partial(_assert_num_queries, pytestconfig)


@pytest.fixture(scope="function")
def django_assert_max_num_queries(pytestconfig):
    return partial(_assert_num_queries, pytestconfig, exact=False)



"Thread-safe in-memory cache backend."
import pickle
import time
from collections import OrderedDict
from threading import Lock

from django.core.cache.backends.base import DEFAULT_TIMEOUT, BaseCache

# Global in-memory store of cache data. Keyed by name, to provide
# multiple named local memory caches.
_caches = {}
_expire_info = {}
_locks = {}


class LocMemCache(BaseCache):
    pickle_protocol = pickle.HIGHEST_PROTOCOL

    def __init__(self, name, params):
        super().__init__(params)
        self._cache = _caches.setdefault(name, OrderedDict())
        self._expire_info = _expire_info.setdefault(name, {})
        self._lock = _locks.setdefault(name, Lock())

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        pickled = pickle.dumps(value, self.pickle_protocol)
        with self._lock:
            if self._has_expired(key):
                self._set(key, pickled, timeout)
                return True
            return False

    def get(self, key, default=None, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        with self._lock:
            if self._has_expired(key):
                self._delete(key)
                return default
            pickled = self._cache[key]
            self._cache.move_to_end(key, last=False)
        return pickle.loads(pickled)

    def _set(self, key, value, timeout=DEFAULT_TIMEOUT):
        if len(self._cache) >= self._max_entries:
            self._cull()
        self._cache[key] = value
        self._cache.move_to_end(key, last=False)
        self._expire_info[key] = self.get_backend_timeout(timeout)

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        pickled = pickle.dumps(value, self.pickle_protocol)
        with self._lock:
            self._set(key, pickled, timeout)

    def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        with self._lock:
            if self._has_expired(key):
                return False
            self._expire_info[key] = self.get_backend_timeout(timeout)
            return True

    def incr(self, key, delta=1, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        with self._lock:
            if self._has_expired(key):
                self._delete(key)
                raise ValueError("Key '%s' not found" % key)
            pickled = self._cache[key]
            value = pickle.loads(pickled)
            new_value = value + delta
            pickled = pickle.dumps(new_value, self.pickle_protocol)
            self._cache[key] = pickled
            self._cache.move_to_end(key, last=False)
        return new_value

    def has_key(self, key, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        with self._lock:
            if self._has_expired(key):
                self._delete(key)
                return False
            return True

    def _has_expired(self, key):
        exp = self._expire_info.get(key, -1)
        return exp is not None and exp <= time.time()

    def _cull(self):
        if self._cull_frequency == 0:
            self._cache.clear()
            self._expire_info.clear()
        else:
            count = len(self._cache) // self._cull_frequency
            for i in range(count):
                key, _ = self._cache.popitem()
                del self._expire_info[key]

    def _delete(self, key):
        try:
            del self._cache[key]
            del self._expire_info[key]
        except KeyError:
            pass

    def delete(self, key, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        with self._lock:
            self._delete(key)

    def clear(self):
        with self._lock:
            self._cache.clear()
            self._expire_info.clear()



from django.db.models import Lookup, Transform
from django.db.models.lookups import Exact

from .search import SearchVector, SearchVectorExact, SearchVectorField


class PostgresSimpleLookup(Lookup):
    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = tuple(lhs_params) + tuple(rhs_params)
        return '%s %s %s' % (lhs, self.operator, rhs), params


class DataContains(PostgresSimpleLookup):
    lookup_name = 'contains'
    operator = '@>'


class ContainedBy(PostgresSimpleLookup):
    lookup_name = 'contained_by'
    operator = '<@'


class Overlap(PostgresSimpleLookup):
    lookup_name = 'overlap'
    operator = '&&'


class HasKey(PostgresSimpleLookup):
    lookup_name = 'has_key'
    operator = '?'
    prepare_rhs = False


class HasKeys(PostgresSimpleLookup):
    lookup_name = 'has_keys'
    operator = '?&'

    def get_prep_lookup(self):
        return [str(item) for item in self.rhs]


class HasAnyKeys(HasKeys):
    lookup_name = 'has_any_keys'
    operator = '?|'


class Unaccent(Transform):
    bilateral = True
    lookup_name = 'unaccent'
    function = 'UNACCENT'


class SearchLookup(SearchVectorExact):
    lookup_name = 'search'

    def process_lhs(self, qn, connection):
        if not isinstance(self.lhs.output_field, SearchVectorField):
            self.lhs = SearchVector(self.lhs)
        lhs, lhs_params = super().process_lhs(qn, connection)
        return lhs, lhs_params


class TrigramSimilar(PostgresSimpleLookup):
    lookup_name = 'trigram_similar'
    operator = '%%'


class JSONExact(Exact):
    can_use_none_as_rhs = True

    def process_rhs(self, compiler, connection):
        result = super().process_rhs(compiler, connection)
        # Treat None lookup values as null.
        return ("'null'", []) if result == ('%s', [None]) else result



from django.conf import settings
from django.utils.translation.trans_real import language_code_re

from . import Error, Tags, register

E001 = Error(
    'You have provided an invalid value for the LANGUAGE_CODE setting.',
    id='translation.E001',
)


@register(Tags.translation)
def check_setting_language_code(app_configs, **kwargs):
    """
    Errors if language code setting is invalid.
    """
    if not language_code_re.match(settings.LANGUAGE_CODE):
        return [E001]
    return []



def get_image_urls(obj, field, sizes={'base': None}):
    images = {}
    for size in sizes:
        img_field = f"{field}_urls"
        try:
            images[size] = (
                getattr(obj, img_field).get(size) if getattr(obj, img_field) else None
            )
        except:
            images[size] = ""
    return images



def luhn_checksum(number):
    def digits_of(n):
        return [int(d) for d in str(n)]

    digits = digits_of(number)
    odd_digits = digits[-1::-2]
    even_digits = digits[-2::-2]
    checksum = 0
    checksum += sum(odd_digits)
    for d in even_digits:
        checksum += sum(digits_of(d * 2))
    return checksum % 10


def calculate_luhn(partial_number):
    """
    Generates the Checksum using Luhn's algorithm
    """
    check_digit = luhn_checksum(int(partial_number) * 10)
    return check_digit if check_digit == 0 else 10 - check_digit



class BaseDatabaseClient:
    """Encapsulate backend-specific methods for opening a client shell."""
    # This should be a string representing the name of the executable
    # (e.g., "psql"). Subclasses must override this.
    executable_name = None

    def __init__(self, connection):
        # connection is an instance of BaseDatabaseWrapper.
        self.connection = connection

    def runshell(self):
        raise NotImplementedError('subclasses of BaseDatabaseClient must provide a runshell() method')



from .. import Provider as BaseProvider


class Provider(BaseProvider):
    """
    A Faker provider for the Cypriot VAT IDs
    """

    vat_id_formats = (
        'CY#########?',
    )

    def vat_id(self):
        """
        Returns a random generated Cypriot Tax ID
        """

        return self.bothify(self.random_element(self.vat_id_formats))



from django.urls import include, path
from scarlet.cms.sites import site
from scarlet import cms

cms.autodiscover()
urlpatterns = [path("admin/", site.urls)]

try:
    from scarlet_blog import blog

    urlpatterns += [
        path(r"", include("scarlet_blog.blog.urls")),
    ]
except ImportError:
    pass



"""
SQLite backend for the sqlite3 module in the standard library.
"""
import datetime
import decimal
import functools
import math
import operator
import re
import statistics
import warnings
from itertools import chain
from sqlite3 import dbapi2 as Database

import pytz

from django.core.exceptions import ImproperlyConfigured
from django.db import utils
from django.db.backends import utils as backend_utils
from django.db.backends.base.base import BaseDatabaseWrapper
from django.utils import timezone
from django.utils.dateparse import parse_datetime, parse_time
from django.utils.duration import duration_microseconds

from .client import DatabaseClient                          # isort:skip
from .creation import DatabaseCreation                      # isort:skip
from .features import DatabaseFeatures                      # isort:skip
from .introspection import DatabaseIntrospection            # isort:skip
from .operations import DatabaseOperations                  # isort:skip
from .schema import DatabaseSchemaEditor                    # isort:skip


def decoder(conv_func):
    """
    Convert bytestrings from Python's sqlite3 interface to a regular string.
    """
    return lambda s: conv_func(s.decode())


def none_guard(func):
    """
    Decorator that returns None if any of the arguments to the decorated
    function are None. Many SQL functions return NULL if any of their arguments
    are NULL. This decorator simplifies the implementation of this for the
    custom functions registered below.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return None if None in args else func(*args, **kwargs)
    return wrapper


def list_aggregate(function):
    """
    Return an aggregate class that accumulates values in a list and applies
    the provided function to the data.
    """
    return type('ListAggregate', (list,), {'finalize': function, 'step': list.append})


def check_sqlite_version():
    if Database.sqlite_version_info < (3, 8, 3):
        raise ImproperlyConfigured('SQLite 3.8.3 or later is required (found %s).' % Database.sqlite_version)


check_sqlite_version()

Database.register_converter("bool", b'1'.__eq__)
Database.register_converter("time", decoder(parse_time))
Database.register_converter("datetime", decoder(parse_datetime))
Database.register_converter("timestamp", decoder(parse_datetime))
Database.register_converter("TIMESTAMP", decoder(parse_datetime))

Database.register_adapter(decimal.Decimal, str)


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'sqlite'
    display_name = 'SQLite'
    # SQLite doesn't actually support most of these types, but it "does the right
    # thing" given more verbose field definitions, so leave them as is so that
    # schema inspection is more useful.
    data_types = {
        'AutoField': 'integer',
        'BigAutoField': 'integer',
        'BinaryField': 'BLOB',
        'BooleanField': 'bool',
        'CharField': 'varchar(%(max_length)s)',
        'DateField': 'date',
        'DateTimeField': 'datetime',
        'DecimalField': 'decimal',
        'DurationField': 'bigint',
        'FileField': 'varchar(%(max_length)s)',
        'FilePathField': 'varchar(%(max_length)s)',
        'FloatField': 'real',
        'IntegerField': 'integer',
        'BigIntegerField': 'bigint',
        'IPAddressField': 'char(15)',
        'GenericIPAddressField': 'char(39)',
        'NullBooleanField': 'bool',
        'OneToOneField': 'integer',
        'PositiveIntegerField': 'integer unsigned',
        'PositiveSmallIntegerField': 'smallint unsigned',
        'SlugField': 'varchar(%(max_length)s)',
        'SmallIntegerField': 'smallint',
        'TextField': 'text',
        'TimeField': 'time',
        'UUIDField': 'char(32)',
    }
    data_type_check_constraints = {
        'PositiveIntegerField': '"%(column)s" >= 0',
        'PositiveSmallIntegerField': '"%(column)s" >= 0',
    }
    data_types_suffix = {
        'AutoField': 'AUTOINCREMENT',
        'BigAutoField': 'AUTOINCREMENT',
    }
    # SQLite requires LIKE statements to include an ESCAPE clause if the value
    # being escaped has a percent or underscore in it.
    # See https://www.sqlite.org/lang_expr.html for an explanation.
    operators = {
        'exact': '= %s',
        'iexact': "LIKE %s ESCAPE '\\'",
        'contains': "LIKE %s ESCAPE '\\'",
        'icontains': "LIKE %s ESCAPE '\\'",
        'regex': 'REGEXP %s',
        'iregex': "REGEXP '(?i)' || %s",
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': "LIKE %s ESCAPE '\\'",
        'endswith': "LIKE %s ESCAPE '\\'",
        'istartswith': "LIKE %s ESCAPE '\\'",
        'iendswith': "LIKE %s ESCAPE '\\'",
    }

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, *, _) should be
    # escaped on database side.
    #
    # Note: we use str.format() here for readability as '%' is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '\\'), '%%', '\%%'), '_', '\_')"
    pattern_ops = {
        'contains': r"LIKE '%%' || {} || '%%' ESCAPE '\'",
        'icontains': r"LIKE '%%' || UPPER({}) || '%%' ESCAPE '\'",
        'startswith': r"LIKE {} || '%%' ESCAPE '\'",
        'istartswith': r"LIKE UPPER({}) || '%%' ESCAPE '\'",
        'endswith': r"LIKE '%%' || {} ESCAPE '\'",
        'iendswith': r"LIKE '%%' || UPPER({}) ESCAPE '\'",
    }

    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor
    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if not settings_dict['NAME']:
            raise ImproperlyConfigured(
                "settings.DATABASES is improperly configured. "
                "Please supply the NAME value.")
        kwargs = {
            'database': settings_dict['NAME'],
            'detect_types': Database.PARSE_DECLTYPES | Database.PARSE_COLNAMES,
            **settings_dict['OPTIONS'],
        }
        # Always allow the underlying SQLite connection to be shareable
        # between multiple threads. The safe-guarding will be handled at a
        # higher level by the `BaseDatabaseWrapper.allow_thread_sharing`
        # property. This is necessary as the shareability is disabled by
        # default in pysqlite and it cannot be changed once a connection is
        # opened.
        if 'check_same_thread' in kwargs and kwargs['check_same_thread']:
            warnings.warn(
                'The `check_same_thread` option was provided and set to '
                'True. It will be overridden with False. Use the '
                '`DatabaseWrapper.allow_thread_sharing` property instead '
                'for controlling thread shareability.',
                RuntimeWarning
            )
        kwargs.update({'check_same_thread': False, 'uri': True})
        return kwargs

    def get_new_connection(self, conn_params):
        conn = Database.connect(**conn_params)
        conn.create_function("django_date_extract", 2, _sqlite_datetime_extract)
        conn.create_function("django_date_trunc", 2, _sqlite_date_trunc)
        conn.create_function("django_datetime_cast_date", 2, _sqlite_datetime_cast_date)
        conn.create_function("django_datetime_cast_time", 2, _sqlite_datetime_cast_time)
        conn.create_function("django_datetime_extract", 3, _sqlite_datetime_extract)
        conn.create_function("django_datetime_trunc", 3, _sqlite_datetime_trunc)
        conn.create_function("django_time_extract", 2, _sqlite_time_extract)
        conn.create_function("django_time_trunc", 2, _sqlite_time_trunc)
        conn.create_function("django_time_diff", 2, _sqlite_time_diff)
        conn.create_function("django_timestamp_diff", 2, _sqlite_timestamp_diff)
        conn.create_function("django_format_dtdelta", 3, _sqlite_format_dtdelta)
        conn.create_function('regexp', 2, _sqlite_regexp)
        conn.create_function('ACOS', 1, none_guard(math.acos))
        conn.create_function('ASIN', 1, none_guard(math.asin))
        conn.create_function('ATAN', 1, none_guard(math.atan))
        conn.create_function('ATAN2', 2, none_guard(math.atan2))
        conn.create_function('CEILING', 1, none_guard(math.ceil))
        conn.create_function('COS', 1, none_guard(math.cos))
        conn.create_function('COT', 1, none_guard(lambda x: 1 / math.tan(x)))
        conn.create_function('DEGREES', 1, none_guard(math.degrees))
        conn.create_function('EXP', 1, none_guard(math.exp))
        conn.create_function('FLOOR', 1, none_guard(math.floor))
        conn.create_function('LN', 1, none_guard(math.log))
        conn.create_function('LOG', 2, none_guard(lambda x, y: math.log(y, x)))
        conn.create_function('LPAD', 3, _sqlite_lpad)
        conn.create_function('MOD', 2, none_guard(math.fmod))
        conn.create_function('PI', 0, lambda: math.pi)
        conn.create_function('POWER', 2, none_guard(operator.pow))
        conn.create_function('RADIANS', 1, none_guard(math.radians))
        conn.create_function('REPEAT', 2, none_guard(operator.mul))
        conn.create_function('REVERSE', 1, none_guard(lambda x: x[::-1]))
        conn.create_function('RPAD', 3, _sqlite_rpad)
        conn.create_function('SIN', 1, none_guard(math.sin))
        conn.create_function('SQRT', 1, none_guard(math.sqrt))
        conn.create_function('TAN', 1, none_guard(math.tan))
        conn.create_aggregate('STDDEV_POP', 1, list_aggregate(statistics.pstdev))
        conn.create_aggregate('STDDEV_SAMP', 1, list_aggregate(statistics.stdev))
        conn.create_aggregate('VAR_POP', 1, list_aggregate(statistics.pvariance))
        conn.create_aggregate('VAR_SAMP', 1, list_aggregate(statistics.variance))
        conn.execute('PRAGMA foreign_keys = ON')
        return conn

    def init_connection_state(self):
        pass

    def create_cursor(self, name=None):
        return self.connection.cursor(factory=SQLiteCursorWrapper)

    def close(self):
        self.validate_thread_sharing()
        # If database is in memory, closing the connection destroys the
        # database. To prevent accidental data loss, ignore close requests on
        # an in-memory db.
        if not self.is_in_memory_db():
            BaseDatabaseWrapper.close(self)

    def _savepoint_allowed(self):
        # When 'isolation_level' is not None, sqlite3 commits before each
        # savepoint; it's a bug. When it is None, savepoints don't make sense
        # because autocommit is enabled. The only exception is inside 'atomic'
        # blocks. To work around that bug, on SQLite, 'atomic' starts a
        # transaction explicitly rather than simply disable autocommit.
        return self.in_atomic_block

    def _set_autocommit(self, autocommit):
        if autocommit:
            level = None
        else:
            # sqlite3's internal default is ''. It's different from None.
            # See Modules/_sqlite/connection.c.
            level = ''
        # 'isolation_level' is a misleading API.
        # SQLite always runs at the SERIALIZABLE isolation level.
        with self.wrap_database_errors:
            self.connection.isolation_level = level

    def disable_constraint_checking(self):
        with self.cursor() as cursor:
            cursor.execute('PRAGMA foreign_keys = OFF')
            # Foreign key constraints cannot be turned off while in a multi-
            # statement transaction. Fetch the current state of the pragma
            # to determine if constraints are effectively disabled.
            enabled = cursor.execute('PRAGMA foreign_keys').fetchone()[0]
        return not bool(enabled)

    def enable_constraint_checking(self):
        self.cursor().execute('PRAGMA foreign_keys = ON')

    def check_constraints(self, table_names=None):
        """
        Check each table name in `table_names` for rows with invalid foreign
        key references. This method is intended to be used in conjunction with
        `disable_constraint_checking()` and `enable_constraint_checking()`, to
        determine if rows with invalid references were entered while constraint
        checks were off.
        """
        if self.features.supports_pragma_foreign_key_check:
            with self.cursor() as cursor:
                if table_names is None:
                    violations = self.cursor().execute('PRAGMA foreign_key_check').fetchall()
                else:
                    violations = chain.from_iterable(
                        cursor.execute('PRAGMA foreign_key_check(%s)' % table_name).fetchall()
                        for table_name in table_names
                    )
                # See https://www.sqlite.org/pragma.html#pragma_foreign_key_check
                for table_name, rowid, referenced_table_name, foreign_key_index in violations:
                    foreign_key = cursor.execute(
                        'PRAGMA foreign_key_list(%s)' % table_name
                    ).fetchall()[foreign_key_index]
                    column_name, referenced_column_name = foreign_key[3:5]
                    primary_key_column_name = self.introspection.get_primary_key_column(cursor, table_name)
                    primary_key_value, bad_value = cursor.execute(
                        'SELECT %s, %s FROM %s WHERE rowid = %%s' % (
                            primary_key_column_name, column_name, table_name
                        ),
                        (rowid,),
                    ).fetchone()
                    raise utils.IntegrityError(
                        "The row in table '%s' with primary key '%s' has an "
                        "invalid foreign key: %s.%s contains a value '%s' that "
                        "does not have a corresponding value in %s.%s." % (
                            table_name, primary_key_value, table_name, column_name,
                            bad_value, referenced_table_name, referenced_column_name
                        )
                    )
        else:
            with self.cursor() as cursor:
                if table_names is None:
                    table_names = self.introspection.table_names(cursor)
                for table_name in table_names:
                    primary_key_column_name = self.introspection.get_primary_key_column(cursor, table_name)
                    if not primary_key_column_name:
                        continue
                    key_columns = self.introspection.get_key_columns(cursor, table_name)
                    for column_name, referenced_table_name, referenced_column_name in key_columns:
                        cursor.execute(
                            """
                            SELECT REFERRING.`%s`, REFERRING.`%s` FROM `%s` as REFERRING
                            LEFT JOIN `%s` as REFERRED
                            ON (REFERRING.`%s` = REFERRED.`%s`)
                            WHERE REFERRING.`%s` IS NOT NULL AND REFERRED.`%s` IS NULL
                            """
                            % (
                                primary_key_column_name, column_name, table_name,
                                referenced_table_name, column_name, referenced_column_name,
                                column_name, referenced_column_name,
                            )
                        )
                        for bad_row in cursor.fetchall():
                            raise utils.IntegrityError(
                                "The row in table '%s' with primary key '%s' has an "
                                "invalid foreign key: %s.%s contains a value '%s' that "
                                "does not have a corresponding value in %s.%s." % (
                                    table_name, bad_row[0], table_name, column_name,
                                    bad_row[1], referenced_table_name, referenced_column_name,
                                )
                            )

    def is_usable(self):
        return True

    def _start_transaction_under_autocommit(self):
        """
        Start a transaction explicitly in autocommit mode.

        Staying in autocommit mode works around a bug of sqlite3 that breaks
        savepoints when autocommit is disabled.
        """
        self.cursor().execute("BEGIN")

    def is_in_memory_db(self):
        return self.creation.is_in_memory_db(self.settings_dict['NAME'])


FORMAT_QMARK_REGEX = re.compile(r'(?<!%)%s')


class SQLiteCursorWrapper(Database.Cursor):
    """
    Django uses "format" style placeholders, but pysqlite2 uses "qmark" style.
    This fixes it -- but note that if you want to use a literal "%s" in a query,
    you'll need to use "%%s".
    """
    def execute(self, query, params=None):
        if params is None:
            return Database.Cursor.execute(self, query)
        query = self.convert_query(query)
        return Database.Cursor.execute(self, query, params)

    def executemany(self, query, param_list):
        query = self.convert_query(query)
        return Database.Cursor.executemany(self, query, param_list)

    def convert_query(self, query):
        return FORMAT_QMARK_REGEX.sub('?', query).replace('%%', '%')


def _sqlite_datetime_parse(dt, tzname=None):
    if dt is None:
        return None
    try:
        dt = backend_utils.typecast_timestamp(dt)
    except (TypeError, ValueError):
        return None
    if tzname is not None:
        dt = timezone.localtime(dt, pytz.timezone(tzname))
    return dt


def _sqlite_date_trunc(lookup_type, dt):
    dt = _sqlite_datetime_parse(dt)
    if dt is None:
        return None
    if lookup_type == 'year':
        return "%i-01-01" % dt.year
    elif lookup_type == 'quarter':
        month_in_quarter = dt.month - (dt.month - 1) % 3
        return '%i-%02i-01' % (dt.year, month_in_quarter)
    elif lookup_type == 'month':
        return "%i-%02i-01" % (dt.year, dt.month)
    elif lookup_type == 'week':
        dt = dt - datetime.timedelta(days=dt.weekday())
        return "%i-%02i-%02i" % (dt.year, dt.month, dt.day)
    elif lookup_type == 'day':
        return "%i-%02i-%02i" % (dt.year, dt.month, dt.day)


def _sqlite_time_trunc(lookup_type, dt):
    if dt is None:
        return None
    try:
        dt = backend_utils.typecast_time(dt)
    except (ValueError, TypeError):
        return None
    if lookup_type == 'hour':
        return "%02i:00:00" % dt.hour
    elif lookup_type == 'minute':
        return "%02i:%02i:00" % (dt.hour, dt.minute)
    elif lookup_type == 'second':
        return "%02i:%02i:%02i" % (dt.hour, dt.minute, dt.second)


def _sqlite_datetime_cast_date(dt, tzname):
    dt = _sqlite_datetime_parse(dt, tzname)
    if dt is None:
        return None
    return dt.date().isoformat()


def _sqlite_datetime_cast_time(dt, tzname):
    dt = _sqlite_datetime_parse(dt, tzname)
    if dt is None:
        return None
    return dt.time().isoformat()


def _sqlite_datetime_extract(lookup_type, dt, tzname=None):
    dt = _sqlite_datetime_parse(dt, tzname)
    if dt is None:
        return None
    if lookup_type == 'week_day':
        return (dt.isoweekday() % 7) + 1
    elif lookup_type == 'week':
        return dt.isocalendar()[1]
    elif lookup_type == 'quarter':
        return math.ceil(dt.month / 3)
    elif lookup_type == 'iso_year':
        return dt.isocalendar()[0]
    else:
        return getattr(dt, lookup_type)


def _sqlite_datetime_trunc(lookup_type, dt, tzname):
    dt = _sqlite_datetime_parse(dt, tzname)
    if dt is None:
        return None
    if lookup_type == 'year':
        return "%i-01-01 00:00:00" % dt.year
    elif lookup_type == 'quarter':
        month_in_quarter = dt.month - (dt.month - 1) % 3
        return '%i-%02i-01 00:00:00' % (dt.year, month_in_quarter)
    elif lookup_type == 'month':
        return "%i-%02i-01 00:00:00" % (dt.year, dt.month)
    elif lookup_type == 'week':
        dt = dt - datetime.timedelta(days=dt.weekday())
        return "%i-%02i-%02i 00:00:00" % (dt.year, dt.month, dt.day)
    elif lookup_type == 'day':
        return "%i-%02i-%02i 00:00:00" % (dt.year, dt.month, dt.day)
    elif lookup_type == 'hour':
        return "%i-%02i-%02i %02i:00:00" % (dt.year, dt.month, dt.day, dt.hour)
    elif lookup_type == 'minute':
        return "%i-%02i-%02i %02i:%02i:00" % (dt.year, dt.month, dt.day, dt.hour, dt.minute)
    elif lookup_type == 'second':
        return "%i-%02i-%02i %02i:%02i:%02i" % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)


def _sqlite_time_extract(lookup_type, dt):
    if dt is None:
        return None
    try:
        dt = backend_utils.typecast_time(dt)
    except (ValueError, TypeError):
        return None
    return getattr(dt, lookup_type)


@none_guard
def _sqlite_format_dtdelta(conn, lhs, rhs):
    """
    LHS and RHS can be either:
    - An integer number of microseconds
    - A string representing a datetime
    """
    try:
        real_lhs = datetime.timedelta(0, 0, lhs) if isinstance(lhs, int) else backend_utils.typecast_timestamp(lhs)
        real_rhs = datetime.timedelta(0, 0, rhs) if isinstance(rhs, int) else backend_utils.typecast_timestamp(rhs)
        if conn.strip() == '+':
            out = real_lhs + real_rhs
        else:
            out = real_lhs - real_rhs
    except (ValueError, TypeError):
        return None
    # typecast_timestamp returns a date or a datetime without timezone.
    # It will be formatted as "%Y-%m-%d" or "%Y-%m-%d %H:%M:%S[.%f]"
    return str(out)


@none_guard
def _sqlite_time_diff(lhs, rhs):
    left = backend_utils.typecast_time(lhs)
    right = backend_utils.typecast_time(rhs)
    return (
        (left.hour * 60 * 60 * 1000000) +
        (left.minute * 60 * 1000000) +
        (left.second * 1000000) +
        (left.microsecond) -
        (right.hour * 60 * 60 * 1000000) -
        (right.minute * 60 * 1000000) -
        (right.second * 1000000) -
        (right.microsecond)
    )


@none_guard
def _sqlite_timestamp_diff(lhs, rhs):
    left = backend_utils.typecast_timestamp(lhs)
    right = backend_utils.typecast_timestamp(rhs)
    return duration_microseconds(left - right)


@none_guard
def _sqlite_regexp(re_pattern, re_string):
    return bool(re.search(re_pattern, str(re_string)))


@none_guard
def _sqlite_lpad(text, length, fill_text):
    if len(text) >= length:
        return text[:length]
    return (fill_text * length)[:length - len(text)] + text


@none_guard
def _sqlite_rpad(text, length, fill_text):
    return (text + fill_text * length)[:length]



import json

from django import forms
from django.db import models

try:
    from ..cms.internal_tags.forms import TaggedRelationFilterForm
except ValueError:
    from cms.internal_tags.forms import TaggedRelationFilterForm

from . import get_asset_model
from .models import AssetBase


class UploadAssetForm(forms.ModelForm):
    """
    Form for handling new asset creation
    """

    class Meta:
        model = get_asset_model()
        fields = ("type", "file", "tags")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["tags"].widget.attrs["class"] = "widget-tags"


class UpdateAssetForm(forms.ModelForm):
    """
    Form for handling asset updates
    """

    class Meta:
        model = get_asset_model()
        fields = ("file", "tags")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["tags"].widget.attrs["class"] = "widget-tags"


class UpdateAssetCropForm(forms.ModelForm):
    """
    Form for handling asset updates including crops
    """

    crops = forms.CharField(required=False, widget=forms.HiddenInput())

    class Meta:
        model = get_asset_model()
        fields = ("file", "tags")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["tags"].widget.attrs["class"] = "widget-tags"

    def clean(self):
        cleaned_data = super().clean()
        if "crops" in cleaned_data.keys():
            if cleaned_data.get("crops") != "":
                try:
                    crops_json = json.loads(cleaned_data.get("crops"))
                    for crop in crops_json:
                        required_keys = ["name", "x", "x2", "y", "y2"]
                        for key in required_keys:
                            if key not in crop.keys():
                                raise forms.ValidationError(
                                    "Could not find required keys for all crops", code="invalid"
                                )

                    cleaned_data["crops"] = crops_json
                except json.JSONDecodeError:
                    raise forms.ValidationError(
                        "Unable to decode crop JSON", code="invalid"
                    )
            else:
                del cleaned_data["crops"]
        return cleaned_data


class AssetFilterForm(TaggedRelationFilterForm):
    """
    Form for handling asset filtering by fields
    """

    choices = [("", "---")] + list(AssetBase.TYPES)
    ftype = forms.ChoiceField(required=False, choices=choices, label="File Type")

    def get_filter(self):
        filter_kwargs = self.get_filter_kwargs()
        args = super().get_filter()

        ftype = filter_kwargs.pop("ftype", None)
        if ftype:
            args.append(models.Q(type=ftype))

        return args


class CropForm(forms.Form):
    x = forms.IntegerField()
    x2 = forms.IntegerField()
    y = forms.IntegerField()
    y2 = forms.IntegerField()

    def check_params(self, v, v2):
        if v != None and v2 != None and v2 < v:
            raise forms.ValidationError("Invalid parameters")

    def clean(self):
        cleaned_data = super().clean()
        self.check_params(cleaned_data.get("x"), cleaned_data.get("x2"))
        self.check_params(cleaned_data.get("y"), cleaned_data.get("y2"))
        return cleaned_data



"""
This module is responsible for generating the check digit and formatting
ISBN numbers.
"""


class ISBN:

    MAX_LENGTH = 13

    def __init__(self, ean=None, group=None, registrant=None, publication=None):
        self.ean = ean
        self.group = group
        self.registrant = registrant
        self.publication = publication


class ISBN13(ISBN):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_digit = self._check_digit()

    def _check_digit(self):
        """ Calculate the check digit for ISBN-13.
        See https://en.wikipedia.org/wiki/International_Standard_Book_Number
        for calculation.
        """
        weights = (1 if x % 2 == 0 else 3 for x in range(12))
        body = ''.join([self.ean, self.group, self.registrant,
                        self.publication])
        remainder = sum(int(b) * w for b, w in zip(body, weights)) % 10
        diff = 10 - remainder
        check_digit = 0 if diff == 10 else diff
        return str(check_digit)

    def format(self, separator=''):
        return separator.join([self.ean, self.group, self.registrant,
                               self.publication, self.check_digit])


class ISBN10(ISBN):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.check_digit = self._check_digit()

    def _check_digit(self):
        """ Calculate the check digit for ISBN-10.
        See https://en.wikipedia.org/wiki/International_Standard_Book_Number
        for calculation.
        """
        weights = range(1, 10)
        body = ''.join([self.group, self.registrant, self.publication])
        remainder = sum(int(b) * w for b, w in zip(body, weights)) % 11
        check_digit = 'X' if remainder == 10 else str(remainder)
        return str(check_digit)

    def format(self, separator=''):
        return separator.join([self.group, self.registrant, self.publication,
                               self.check_digit])



from django.db import models

from scarlet.versioning.fields import FKToVersion
from scarlet.versioning.models import Cloneable, VersionView

from scarlet.pagebuilder import settings, base_models

"""
This file can be copied into a project's app folder as models.py for
default implementation of Pagebuilder.
"""


class Page(VersionView, base_models.BasePage):
    _clone_related = [
        "page_hero_modules",
        "page_two_column_modules",
        "page_icon_list_modules",
        "page_header_modules",
        "page_faq_modules",
        "page_location_modules",
        "page_gallery_modules",
        "page_carousel_modules",
    ]

    @property
    def site_name(self):
        return settings.SITE_NAME


class HeroModule(base_models.BaseHeroModule):
    page = FKToVersion(
        settings.PAGE_MODEL, related_name="page_hero_modules", blank=True, null=True
    )
    styles = (("left", "left"), ("right", "right"), ("center", "center"))
    behaviors = (("horizontal", "horizontal"), ("vertical", "vertical"))


class TwoColumnModule(base_models.BaseTwoColumnModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="page_two_column_modules",
        blank=True,
        null=True,
    )


class IconListModule(base_models.BaseIconListModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="page_icon_list_modules",
        blank=True,
        null=True,
    )


class IconListItem(base_models.BaseIconListItem):
    module = models.ForeignKey(
        settings.ICON_LIST_MODEL, on_delete=models.CASCADE, blank=True, null=True
    )


class HeaderModule(base_models.BaseHeaderModule):
    page = FKToVersion(
        settings.PAGE_MODEL, related_name="page_header_modules", blank=True, null=True,
    )
    styles = (("left", "left"), ("right", "right"), ("center", "center"))
    behaviors = (("horizontal", "horizontal"), ("vertical", "vertical"))


class FAQModule(base_models.BaseFAQModule):
    page = FKToVersion(
        settings.PAGE_MODEL, related_name="page_faq_modules", blank=True, null=True
    )


class FAQItem(base_models.BaseFAQItem):
    module = models.ForeignKey(
        settings.FAQ_MODULE_MODEL, on_delete=models.CASCADE, blank=True, null=True
    )


class LocationModule(base_models.BaseLocationModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="page_location_modules",
        blank=True,
        null=True,
    )


class LocationItem(base_models.BaseLocationItem):
    module = models.ForeignKey(
        settings.LOCATION_MODULE_MODEL, on_delete=models.CASCADE, blank=True, null=True,
    )


class ImageGalleryModule(base_models.BaseImageGalleryModule):
    page = FKToVersion(
        settings.PAGE_MODEL, related_name="page_gallery_modules", blank=True, null=True,
    )


class GalleryImage(base_models.BaseGalleryImage):
    module = models.ForeignKey(
        settings.GALLERY_MODULE_MODEL, on_delete=models.CASCADE, blank=True, null=True,
    )


class CarouselModule(base_models.BaseCarouselModule):
    page = FKToVersion(
        settings.PAGE_MODEL,
        related_name="page_carousel_modules",
        blank=True,
        null=True,
    )


class CarouselItem(base_models.BaseCarouselItem):
    module = models.ForeignKey(
        settings.CAROUSEL_MODULE_MODEL, on_delete=models.CASCADE, blank=True, null=True,
    )



from django import forms
from django.forms.models import inlineformset_factory
from scarlet.cms.forms import LazyFormSetFactory
from .settings import (
    MODULE_TYPES,
    HERO_MODULE,
    HEADER_MODULE,
    FAQ_MODULE,
    TWO_COLUMN_MODULE,
    ICON_LIST_MODULE,
    LOCATION_MODULE,
    GALLERY_MODULE,
    CAROUSEL_MODULE,
)
from . import (
    get_page_model,
    get_hero_model,
    get_two_column_model,
    get_icon_list_model,
    get_icon_item_model,
    get_header_model,
    get_faq_model,
    get_faq_item_model,
    get_location_model,
    get_location_item_model,
    get_image_gallery_model,
    get_gallery_item_model,
    get_carousel_model,
    get_carousel_item_model,
)

Page = get_page_model()
HeroModule = get_hero_model()
HeaderModule = get_header_model()
TwoColumnModule = get_two_column_model()
IconListModule = get_icon_list_model()
IconListItem = get_icon_item_model()
FAQModule = get_faq_model()
FAQItem = get_faq_item_model()
LocationModule = get_location_model()
LocationItem = get_location_item_model()
ImageGalleryModule = get_image_gallery_model()
GalleryImage = get_gallery_item_model()
CarouselModule = get_carousel_model()
CarouselItem = get_carousel_item_model()


class PageForm(forms.ModelForm):
    class Meta:
        model = Page
        fields = (
            "title",
            "slug",
            "internal_name",
        )


class EmptyPageForm(forms.ModelForm):
    class Meta:
        model = Page
        exclude = (
            "title",
            "slug",
            "description",
            "keywords",
            "og_image",
            "og_image_alt_text",
        )


class HeroModuleForm(forms.ModelForm):
    class Meta:
        model = HeroModule
        fields = "__all__"


class TwoColumnModuleForm(forms.ModelForm):
    class Meta:
        model = TwoColumnModule
        fields = "__all__"


class HeaderModuleForm(forms.ModelForm):
    class Meta:
        model = HeaderModule
        fields = "__all__"


class IconListModuleForm(forms.ModelForm):
    class Meta:
        model = IconListModule
        fields = "__all__"


class IconListItemForm(forms.ModelForm):
    class Meta:
        model = IconListItem
        fields = "__all__"


class FAQModuleForm(forms.ModelForm):
    class Meta:
        model = FAQModule
        fields = "__all__"


class FAQItemForm(forms.ModelForm):
    class Meta:
        model = FAQItem
        fields = "__all__"


class LocationModuleForm(forms.ModelForm):
    class Meta:
        model = LocationModule
        fields = "__all__"


class LocationItemForm(forms.ModelForm):
    class Meta:
        model = LocationItem
        fields = "__all__"


class ImageGalleryModuleForm(forms.ModelForm):
    class Meta:
        model = ImageGalleryModule
        fields = "__all__"


class GalleryImageForm(forms.ModelForm):
    class Meta:
        model = GalleryImage
        fields = "__all__"


class CarouselModuleForm(forms.ModelForm):
    class Meta:
        model = CarouselModule
        fields = "__all__"


class CarouselItemForm(forms.ModelForm):
    class Meta:
        model = CarouselItem
        fields = "__all__"


class HeroInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(HeroInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            HeroModule,
            can_order=False,
            can_delete=True,
            form=HeroModuleForm,
            fk_name="page",
        )


class TwoColumnInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(TwoColumnInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            TwoColumnModule,
            can_order=False,
            can_delete=True,
            form=TwoColumnModuleForm,
            fk_name="page",
        )


class IconListInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(IconListInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            IconListModule,
            can_order=False,
            can_delete=True,
            form=IconListModuleForm,
            fk_name="page",
        )


class IconItemInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(IconItemInlineFormset, self).__init__(
            inlineformset_factory,
            IconListModule,
            IconListItem,
            can_order=False,
            can_delete=True,
            form=IconListItemForm,
            fk_name="module",
        )


class HeaderInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(HeaderInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            HeaderModule,
            can_order=False,
            can_delete=True,
            form=HeaderModuleForm,
            fk_name="page",
        )


class FAQInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(FAQInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            FAQModule,
            can_order=False,
            can_delete=True,
            form=FAQModuleForm,
            fk_name="page",
        )


class FAQItemInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(FAQItemInlineFormset, self).__init__(
            inlineformset_factory,
            FAQModule,
            FAQItem,
            can_order=False,
            can_delete=True,
            form=FAQItemForm,
            fk_name="module",
        )


class LocationInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(LocationInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            LocationModule,
            can_order=False,
            can_delete=True,
            form=LocationModuleForm,
            fk_name="page",
        )


class LocationItemInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(LocationItemInlineFormset, self).__init__(
            inlineformset_factory,
            LocationModule,
            LocationItem,
            can_order=False,
            can_delete=True,
            form=LocationItemForm,
            fk_name="module",
        )


class ImageGalleryInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(ImageGalleryInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            ImageGalleryModule,
            can_order=False,
            can_delete=True,
            form=ImageGalleryModuleForm,
            fk_name="page",
        )


class GalleryImageInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(GalleryImageInlineFormset, self).__init__(
            inlineformset_factory,
            ImageGalleryModule,
            GalleryImage,
            can_order=False,
            can_delete=True,
            form=GalleryImageForm,
            fk_name="module",
        )


class CarouselInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(CarouselInlineFormset, self).__init__(
            inlineformset_factory,
            Page,
            CarouselModule,
            can_order=False,
            can_delete=True,
            form=CarouselModuleForm,
            fk_name="page",
        )


class CarouselItemInlineFormset(LazyFormSetFactory):
    def __init__(self):
        super(CarouselItemInlineFormset, self).__init__(
            inlineformset_factory,
            CarouselModule,
            CarouselItem,
            can_order=False,
            can_delete=True,
            form=CarouselItemForm,
            fk_name="module",
        )


PAGE_EDIT_FIELDSET = (("Content Modules", {"fields": (),}),)

PAGE_EDIT_FORMSETS = {
    HERO_MODULE: HeroInlineFormset(),
    TWO_COLUMN_MODULE: TwoColumnInlineFormset(),
    HEADER_MODULE: HeaderInlineFormset(),
    ICON_LIST_MODULE: IconListInlineFormset(),
    FAQ_MODULE: FAQInlineFormset(),
    LOCATION_MODULE: LocationInlineFormset(),
    GALLERY_MODULE: ImageGalleryInlineFormset(),
    CAROUSEL_MODULE: CarouselInlineFormset(),
}

MODULE_COMBINED_FORMSET = {
    "title": "Modules",
    "keys": MODULE_TYPES,
}



from . import get_asset_model, get_image_cropper, settings


def optional_celery(**kparms):
    name = kparms.pop("name", None)

    def wrapped(func):
        def inner(*args, **kw):
            return func(*args, **kw)

        return inner

    if settings.USE_CELERY_DECORATOR:
        from celery import task

        wrapper = task(**kparms)

    elif settings.CELERY:
        wrapper = settings.CELERY.task(**kparms)
    else:
        wrapper = wrapped
    return wrapper


@optional_celery(name="assets_ensure_crops")
def ensure_crops(asset_id, *required_crops, **kwargs):
    asset = kwargs.pop("asset", None)
    if not asset or asset_id:
        asset = get_asset_model().objects.get(pk=asset_id)

    required_crops = set(required_crops).union(
        set(get_image_cropper().required_crops())
    )

    crops = set(asset.imagedetail_set.all().values_list("name", flat=True))
    needed = required_crops.difference(crops)
    length = len(needed)
    detail_mod = asset.imagedetail_set.model
    for i, size in enumerate(needed):
        last = i == (length - 1)
        spec = get_image_cropper().create_crop(size, asset.file)
        detail_mod.save_crop_spec(asset, spec, update_version=last)


@optional_celery(name="assets_reset_crops")
def reset_crops(asset_id, asset=None, **kwargs):
    if not asset or asset_id:
        asset = get_asset_model().objects.get(pk=asset_id)

    crops = set(asset.imagedetail_set.values_list("name", flat=True))
    crops = crops.union(set(get_image_cropper().required_crops()))
    length = len(crops)
    detail_mod = asset.imagedetail_set.model
    for i, size in enumerate(crops):
        last = i == (length - 1)
        spec = get_image_cropper().create_crop(size, asset.file)
        detail_mod.save_crop_spec(asset, spec, update_version=last)



from .. import Provider as PersonProvider


class Provider(PersonProvider):
    formats_male = (
        '{{first_name_male}} {{last_name}}',
    )
    formats_female = (
        '{{first_name_female}} {{last_name}}',
    )

    formats = formats_male + formats_female

    # Source: 2012 Voters List.
    # Obtained from http://mashasada.me/en/chamotvirtva
    first_names_male = (
        'ავთანდილ', 'აკაკი', 'ალექს', 'ალექსანდრე', 'ამირან', 'ანზორ', 'არმენ',
        'არტურ', 'არჩილ', 'ასლან', 'ბადრი', 'ბესარიონ', 'ბესიკ', 'ბექა',
        'ბიძინა', 'ბონდო', 'ბორის', 'გელა', 'გენადი', 'გია', 'გიგა', 'გიგლა',
        'გივი', 'გიორგი', 'გოგა', 'გოგი', 'გოგიტა', 'გოდერძი', 'გოჩა', 'გრიგოლ',
        'გურამ', 'დავით', 'დათო', 'დემურ', 'დიმიტრი', 'ედუარდ', 'ელგუჯა',
        'ემზარ', 'ვალერი', 'ვალერიან', 'ვანო', 'ვაჟა', 'ვასილ', 'ვახტანგ',
        'ვახტანგ', 'ვეფხვია', 'ვიქტორ', 'ვლადიმერ', 'ზაზა', 'ზაურ', 'ზაქარია',
        'ზვიად', 'ზურაბ', 'თამაზ', 'თეიმურაზ', 'თემურ', 'თენგიზ', 'თორნიკე',
        'იაგო', 'ივანე', 'ილია', 'იოსებ', 'ირაკლი', 'იური', 'კარლო', 'კახა',
        'კახაბერ', 'კობა', 'კონსტანტინე', 'ლაშა', 'ლევან', 'მალხაზ', 'მამუკა',
        'მერაბ', 'მინდია', 'მირიან', 'მიხეილ', 'მიხეილ', 'მურთაზ', 'მურმან',
        'ნიკა', 'ნიკოლოზ', 'ნოდარ', 'ნუგზარ', 'ნუკრი', 'ოთარ', 'ომარ', 'პაატა',
        'პავლე', 'პეტრე', 'რამაზ', 'რატი', 'რევაზ', 'რეზო', 'რობერტ', 'როინ',
        'როლანდ', 'რომან', 'სერგო', 'სიმონ', 'სოსო', 'ტარიელ', 'უშანგი', 'უჩა',
        'შალვა', 'შაქრო', 'შოთა', 'ხვიჩა', 'ჯაბა', 'ჯემალ', 'ჯონი', 'ჯუმბერ',
    )

    # Source: 2012 Voters List.
    # Obtained from http://mashasada.me/en/chamotvirtva
    first_names_female = (
        'აზა', 'აიდა', 'ალა', 'ანა', 'ანი', 'ანიკო', 'ანნა', 'ანჟელა', 'ასმათ',
        'ბელა', 'გალინა', 'გვანცა', 'გიული', 'გუგული', 'გულიკო', 'გულნარა',
        'დალი', 'დარეჯან', 'დიანა', 'დოდო', 'ევგენია', 'ეთერ', 'ეკა',
        'ეკატერინე', 'ელენა', 'ელენე', 'ელზა', 'ელიზა', 'ელისო', 'ელმირა',
        'ემა', 'ვალენტინა', 'ვარდო', 'ვენერა', 'ვერა', 'ვერიკო', 'ვიოლეტა',
        'ზაირა', 'ზინა', 'ზოია', 'თათია', 'თამარ', 'თამარა', 'თამარი', 'თამთა',
        'თამილა', 'თამუნა', 'თეა', 'თეონა', 'თინა', 'თინათინ', 'ია', 'იამზე',
        'იზა', 'იზოლდა', 'ინგა', 'ინეზა', 'ირინა', 'ირინე', 'ირმა', 'კარინე',
        'კლარა', 'ლალი', 'ლამარა', 'ლამზირა', 'ლანა', 'ლარისა', 'ლეილა', 'ლელა',
        'ლენა', 'ლია', 'ლიანა', 'ლიდა', 'ლიზა', 'ლიკა', 'ლილი', 'ლუბა',
        'ლუდმილა', 'ლუიზა', 'მაგდა', 'მადონა', 'მაია', 'მაკა', 'მანანა',
        'მარგალიტა', 'მარი', 'მარია', 'მარიამ', 'მარიკა', 'მარინა', 'მარინე',
        'მარო', 'მაყვალა', 'მეგი', 'მედეა', 'მედიკო', 'მერი', 'მზია',
        'მთვარისა', 'მირანდა', 'ნადეჯდა', 'ნადია', 'ნაზი', 'ნაზიბროლა',
        'ნაზიკო', 'ნათელა', 'ნათია', 'ნაირა', 'ნანა', 'ნანი', 'ნანული',
        'ნარგიზა', 'ნატალია', 'ნატო', 'ნელი', 'ნესტან', 'ნინელი', 'ნინო',
        'ნონა', 'ნორა', 'ნუნუ', 'ნუცა', 'ოლია', 'ოლღა', 'ჟანა', 'ჟენია',
        'ჟუჟუნა', 'რიმა', 'რიტა', 'როზა', 'რუსუდან', 'რუსუდან', 'სალომე',
        'სვეტლანა', 'სონია', 'სოფია', 'სოფიკო', 'სოფიო', 'სულიკო', 'სუსანა',
        'ტატიანა', 'ფატი', 'ფატიმა', 'ფიქრია', 'ქეთევან', 'ქეთინო', 'ქეთო',
        'ქრისტინე', 'შორენა', 'ციალა', 'ცირა', 'ცისანა', 'ციური', 'ციცინო',
        'ხათუნა', 'ხატია', 'ჯულიეტა',
    )

    first_names = first_names_male + first_names_female

    # Source: 2012 Voters List.
    # Obtained from http://mashasada.me/en/chamotvirtva
    last_names = (
        'აბაშიძე', 'აბდულაევი', 'აბესაძე', 'აბრამიშვილი', 'აბულაძე', 'ადამია',
        'ადეიშვილი', 'ადუაშვილი', 'ავალიანი', 'ალადაშვილი', 'ალანია', 'ალიევა',
        'ალიევი', 'ანდღულაძე', 'ანთაძე', 'არაბიძე', 'არაბული', 'არველაძე',
        'არჩვაძე', 'ასათიანი', 'ასანიძე', 'აფციაური', 'ახალაია', 'ახალაძე',
        'ახვლედიანი', 'ახობაძე', 'ბაკურაძე', 'ბალიაშვილი', 'ბარამიძე',
        'ბარბაქაძე', 'ბასილაშვილი', 'ბასილაძე', 'ბაქრაძე', 'ბახტაძე',
        'ბეგიაშვილი', 'ბენდელიანი', 'ბენიძე', 'ბეჟანიშვილი', 'ბეჟანიძე',
        'ბერაია', 'ბერაძე', 'ბერიანიძე', 'ბერიაშვილი', 'ბერიშვილი', 'ბერიძე',
        'ბერუაშვილი', 'ბერულავა', 'ბერძენიშვილი', 'ბექაური', 'ბიბილაშვილი',
        'ბიგვავა', 'ბიწაძე', 'ბლიაძე', 'ბოკუჩავა', 'ბოლქვაძე', 'ბოჭორიშვილი',
        'ბრეგაძე', 'ბრეგვაძე', 'ბუკია', 'ბურდული', 'ბურჯანაძე', 'ბუჩუკური',
        'ბუცხრიკიძე', 'გაბაიძე', 'გაბედავა', 'გაბელია', 'გაბრიჭიძე', 'გაბუნია',
        'გაგნიძე', 'გაგუა', 'გაფრინდაშვილი', 'გაჩეჩილაძე', 'გაჯიევა', 'გაჯიევი',
        'გეგეშიძე', 'გელაშვილი', 'გელაძე', 'გვაზავა', 'გვასალია', 'გველესიანი',
        'გვენეტაძე', 'გიგაური', 'გიორგაძე', 'გიორგობიანი', 'გობეჯიშვილი',
        'გოგალაძე', 'გოგია', 'გოგიაშვილი', 'გოგინაშვილი', 'გოგიტიძე',
        'გოგიშვილი', 'გოგიჩაიშვილი', 'გოგიჩაშვილი', 'გოგიძე', 'გოგოლაძე',
        'გოგოხია', 'გოგსაძე', 'გოგუა', 'გოგუაძე', 'გორგაძე', 'გორგილაძე',
        'გორგოძე', 'გოცირიძე', 'გრიგალაშვილი', 'გრიგორიანი', 'გრძელიშვილი',
        'გულიაშვილი', 'გულუა', 'გუმბერიძე', 'გურგენიძე', 'გურეშიძე', 'გურული',
        'გუჯაბიძე', 'დავითაშვილი', 'დავითაძე', 'დალაქიშვილი', 'დანელია',
        'დარბაიძე', 'დევაძე', 'დევიძე', 'დემეტრაშვილი', 'დვალი', 'დვალიშვილი',
        'დიასამიძე', 'დოლიძე', 'დუმბაძე', 'ელბაქიძე', 'ელიზბარაშვილი',
        'ენუქიძე', 'ვალიევა', 'ვალიევი', 'ვარდოსანიძე', 'ვარშანიძე', 'ვასაძე',
        'ვაშაკიძე', 'ვაშაყმაძე', 'ვეკუა', 'ზარანდია', 'ზარიძე', 'ზარქუა',
        'ზედგინიძე', 'ზოიძე', 'ზურაბაშვილი', 'თაბაგარი', 'თავართქილაძე',
        'თავაძე', 'თედორაძე', 'თევზაძე', 'თოდუა', 'თოლორდავა', 'თოფურია',
        'თურმანიძე', 'იაშვილი', 'ილურიძე', 'იმერლიშვილი', 'იმნაძე', 'ინასარიძე',
        'იობიძე', 'ირემაშვილი', 'ირემაძე', 'ისაევი', 'კაზარიანი', 'კაკაბაძე',
        'კაკაურიძე', 'კაკულია', 'კალანდაძე', 'კალანდია', 'კანდელაკი',
        'კაპანაძე', 'კარაპეტიანი', 'კაცაძე', 'კაციტაძე', 'კაჭარავა', 'კახაძე',
        'კახიძე', 'კევლიშვილი', 'კეკელიძე', 'კერესელიძე', 'კვანტალიანი',
        'კვარაცხელია', 'კვაჭაძე', 'კვერნაძე', 'კვინიკაძე', 'კვირიკაშვილი',
        'კვირკველია', 'კიკვაძე', 'კიკნაძე', 'კილასონია', 'კილაძე',
        'კინწურაშვილი', 'კირვალიძე', 'კირთაძე', 'კირკიტაძე', 'კობაიძე',
        'კობახიძე', 'კობერიძე', 'კობიაშვილი', 'კონცელიძე', 'კოპალიანი',
        'კოპაძე', 'კოხრეიძე', 'კუბლაშვილი', 'კუპატაძე', 'კუპრაშვილი',
        'კუპრეიშვილი', 'კურტანიძე', 'კუჭავა', 'კუჭუხიძე', 'კუხიანიძე', 'ლაბაძე',
        'ლაგვილავა', 'ლატარია', 'ლაცაბიძე', 'ლეჟავა', 'ლიპარტელიანი',
        'ლობჟანიძე', 'ლობჯანიძე', 'ლოლაძე', 'ლომაძე', 'ლომთაძე', 'ლომიძე',
        'ლომსაძე', 'ლორთქიფანიძე', 'ლურსმანაშვილი', 'მაზმიშვილი', 'მათიაშვილი',
        'მაისურაძე', 'მამალაძე', 'მამულაშვილი', 'მამულაძე', 'მანაგაძე',
        'მარგველაშვილი', 'მაღლაკელიძე', 'მაღრაძე', 'მაჩიტიძე', 'მაჭავარიანი',
        'მაჭარაშვილი', 'მახათაძე', 'მახარაშვილი', 'მახარაძე', 'მგალობლიშვილი',
        'მგელაძე', 'მეგრელიშვილი', 'მელაძე', 'მელიქიშვილი', 'მელიქიძე',
        'მელქაძე', 'მერაბიშვილი', 'მერებაშვილი', 'მესხი', 'მეტრეველი',
        'მეფარიშვილი', 'მიმინოშვილი', 'მინდიაშვილი', 'მიქაბერიძე', 'მიქავა',
        'მიქაძე', 'მიქელაძე', 'მოდებაძე', 'მოსიაშვილი', 'მჟავანაძე', 'მუმლაძე',
        'მუსაევა', 'მუსაევი', 'მუსტაფაევა', 'მუსტაფაევი', 'მუშკუდიანი',
        'მღებრიშვილი', 'მჭედლიშვილი', 'მჭედლიძე', 'ნაბიევი', 'ნადირაშვილი',
        'ნადირაძე', 'ნაკაშიძე', 'ნარიმანიძე', 'ნასყიდაშვილი', 'ნატროშვილი',
        'ნაცვლიშვილი', 'ნაჭყებია', 'ნებიერიძე', 'ნემსაძე', 'ნეფარიძე',
        'ნიკოლაიშვილი', 'ნიკოლეიშვილი', 'ნიჟარაძე', 'ნიქაბაძე', 'ნოზაძე',
        'ნუცუბიძე', 'ოთარაშვილი', 'ონიანი', 'ოქროპირიძე', 'ოქრუაშვილი',
        'პავლიაშვილი', 'პაპავა', 'პაპაშვილი', 'პაპიაშვილი', 'პაპიძე',
        'პაპუაშვილი', 'პატარაია', 'პეტრიაშვილი', 'პეტროსიანი', 'ჟვანია',
        'ჟორჟოლიანი', 'ჟღენტი', 'რაზმაძე', 'რამიშვილი', 'რევაზიშვილი',
        'რეხვიაშვილი', 'რობაქიძე', 'როგავა', 'როსტიაშვილი', 'რუხაძე',
        'სალუქვაძე', 'სამუშია', 'სამხარაძე', 'სანიკიძე', 'სარალიძე',
        'სარქისიანი', 'საჯაია', 'სეხნიაშვილი', 'სვანიძე', 'სილაგაძე',
        'სიმონიშვილი', 'სირაძე', 'სირბილაძე', 'სიჭინავა', 'სიხარულიძე',
        'სოფრომაძე', 'სულაბერიძე', 'სურმანიძე', 'სხირტლაძე', 'ტაბატაძე',
        'ტაკიძე', 'ტალახაძე', 'ტურაშვილი', 'ტუღუში', 'ტყებუჩავა', 'ტყემალაძე',
        'ტყეშელაშვილი', 'უგრეხელიძე', 'ურუშაძე', 'ფანცულაია', 'ფარტენაძე',
        'ფარცვანია', 'ფეიქრიშვილი', 'ფერაძე', 'ფირცხალავა', 'ფიფია',
        'ფიცხელაური', 'ფოფხაძე', 'ფუტკარაძე', 'ფხაკაძე', 'ფხალაძე', 'ქავთარაძე',
        'ქათამაძე', 'ქამადაძე', 'ქანთარია', 'ქარდავა', 'ქართველიშვილი',
        'ქარჩავა', 'ქარცივაძე', 'ქაჯაია', 'ქევხიშვილი', 'ქირია', 'ქობალია',
        'ქობულაძე', 'ქორიძე', 'ქუთათელაძე', 'ქურდაძე', 'ღამბაშიძე',
        'ღარიბაშვილი', 'ღვინიაშვილი', 'ღვინჯილია', 'ღლონტი', 'ღონღაძე',
        'ღურწკაია', 'ყაველაშვილი', 'ყიფიანი', 'ყიფშიძე', 'ყოლბაია', 'ყურაშვილი',
        'შავაძე', 'შათირიშვილი', 'შაინიძე', 'შალამბერიძე', 'შამათავა',
        'შამუგია', 'შანიძე', 'შარაშიძე', 'შარიქაძე', 'შელია', 'შენგელია',
        'შეყილაძე', 'შველიძე', 'შონია', 'შუბითიძე', 'შუკაკიძე', 'ჩადუნელი',
        'ჩაკვეტაძე', 'ჩაფიძე', 'ჩაჩანიძე', 'ჩაჩუა', 'ჩინჩალაძე', 'ჩიქოვანი',
        'ჩიხლაძე', 'ჩოხელი', 'ჩუბინიძე', 'ჩხაიძე', 'ჩხარტიშვილი', 'ჩხეიძე',
        'ჩხიკვაძე', 'ცაავა', 'ცანავა', 'ცარციძე', 'ცერცვაძე', 'ცეცხლაძე',
        'ცინცაძე', 'ცირეკიძე', 'ცომაია', 'ცუცქირიძე', 'ცხადაძე', 'ძიძიგური',
        'ძნელაძე', 'წერეთელი', 'წიკლაური', 'წიქარიშვილი', 'წულაია', 'წულუკიძე',
        'წურწუმია', 'ჭანტურია', 'ჭანტურიძე', 'ჭაღალიძე', 'ჭეიშვილი', 'ჭელიძე',
        'ჭითანავა', 'ჭინჭარაული', 'ჭიღლაძე', 'ჭიჭინაძე', 'ჭკადუა', 'ჭუმბურიძე',
        'ხალვაში', 'ხარაბაძე', 'ხარაზიშვილი', 'ხარაიშვილი', 'ხარატიშვილი',
        'ხარებავა', 'ხარშილაძე', 'ხარჩილავა', 'ხატიაშვილი', 'ხაჩატურიანი',
        'ხაჩიძე', 'ხაჭაპურიძე', 'ხელაძე', 'ხეცურიანი', 'ხვედელიძე', 'ხვიჩია',
        'ხიზანიშვილი', 'ხიმშიაშვილი', 'ხმალაძე', 'ხოზრევანიძე', 'ხორავა',
        'ხურცილავა', 'ხურციძე', 'ხუციშვილი', 'ჯავახიშვილი', 'ჯაიანი',
        'ჯალაღონია', 'ჯანაშია', 'ჯანელიძე', 'ჯანიაშვილი', 'ჯანჯღავა',
        'ჯაფარიძე', 'ჯიბლაძე', 'ჯინჭარაძე', 'ჯიქია', 'ჯიშკარიანი', 'ჯოხაძე',
        'ჯოჯუა', 'ჯღარკავა',
    )



import uuid
from operator import attrgetter

from django import VERSION
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import connections, models, router
from django.db.models import signals
from django.db.models.fields.related import (
    ManyToManyRel,
    OneToOneRel,
    RelatedField,
    lazy_related_operation,
)
from django.db.models.query_utils import PathInfo
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from taggit.forms import TagField
from taggit.models import (
    CommonGenericTaggedItemBase,
    GenericUUIDTaggedItemBase,
    TaggedItem,
)
from taggit.utils import require_instance_manager


class ExtraJoinRestriction:
    """
    An extra restriction used for contenttype restriction in joins.
    """

    contains_aggregate = False

    def __init__(self, alias, col, content_types):
        self.alias = alias
        self.col = col
        self.content_types = content_types

    def as_sql(self, compiler, connection):
        qn = compiler.quote_name_unless_alias
        if len(self.content_types) == 1:
            extra_where = "{}.{} = %s".format(qn(self.alias), qn(self.col))
        else:
            extra_where = "{}.{} IN ({})".format(
                qn(self.alias), qn(self.col), ",".join(["%s"] * len(self.content_types))
            )
        return extra_where, self.content_types

    def relabel_aliases(self, change_map):
        self.alias = change_map.get(self.alias, self.alias)

    def clone(self):
        return type(self)(self.alias, self.col, self.content_types[:])


class _TaggableManager(models.Manager):
    def __init__(self, through, model, instance, prefetch_cache_name):
        super().__init__()
        self.through = through
        self.model = model
        self.instance = instance
        self.prefetch_cache_name = prefetch_cache_name

    def is_cached(self, instance):
        return self.prefetch_cache_name in instance._prefetched_objects_cache

    def get_queryset(self, extra_filters=None):
        try:
            return self.instance._prefetched_objects_cache[self.prefetch_cache_name]
        except (AttributeError, KeyError):
            kwargs = extra_filters if extra_filters else {}
            return self.through.tags_for(self.model, self.instance, **kwargs)

    def get_prefetch_queryset(self, instances, queryset=None):
        if queryset is not None:
            raise ValueError("Custom queryset can't be used for this lookup.")

        instance = instances[0]
        db = self._db or router.db_for_read(type(instance), instance=instance)

        fieldname = (
            "object_id"
            if issubclass(self.through, CommonGenericTaggedItemBase)
            else "content_object"
        )
        fk = self.through._meta.get_field(fieldname)
        query = {
            "%s__%s__in"
            % (self.through.tag_relname(), fk.name): {
                obj._get_pk_val() for obj in instances
            }
        }
        join_table = self.through._meta.db_table
        source_col = fk.column
        connection = connections[db]
        qn = connection.ops.quote_name
        qs = (
            self.get_queryset(query)
            .using(db)
            .extra(
                select={
                    "_prefetch_related_val": "{}.{}".format(
                        qn(join_table), qn(source_col)
                    )
                }
            )
        )

        if issubclass(self.through, GenericUUIDTaggedItemBase):

            def uuid_rel_obj_attr(v):
                value = attrgetter("_prefetch_related_val")(v)
                if value is not None and not isinstance(value, uuid.UUID):
                    input_form = "int" if isinstance(value, int) else "hex"
                    value = uuid.UUID(**{input_form: value})
                return value

            rel_obj_attr = uuid_rel_obj_attr
        else:
            rel_obj_attr = attrgetter("_prefetch_related_val")

        if VERSION < (2, 0):
            return (
                qs,
                rel_obj_attr,
                lambda obj: obj._get_pk_val(),
                False,
                self.prefetch_cache_name,
            )
        else:
            return (
                qs,
                rel_obj_attr,
                lambda obj: obj._get_pk_val(),
                False,
                self.prefetch_cache_name,
                False,
            )

    def _lookup_kwargs(self):
        return self.through.lookup_kwargs(self.instance)

    @require_instance_manager
    def add(self, *tags, **kwargs):
        tag_kwargs = kwargs.pop("tag_kwargs", {})

        db = router.db_for_write(self.through, instance=self.instance)

        tag_objs = self._to_tag_model_instances(tags, tag_kwargs)
        new_ids = {t.pk for t in tag_objs}

        # NOTE: can we hardcode 'tag_id' here or should the column name be got
        # dynamically from somewhere?
        vals = (
            self.through._default_manager.using(db)
            .values_list("tag_id", flat=True)
            .filter(**self._lookup_kwargs())
        )

        new_ids = new_ids - set(vals)

        signals.m2m_changed.send(
            sender=self.through,
            action="pre_add",
            instance=self.instance,
            reverse=False,
            model=self.through.tag_model(),
            pk_set=new_ids,
            using=db,
        )

        for tag in tag_objs:
            self.through._default_manager.using(db).get_or_create(
                tag=tag, **self._lookup_kwargs()
            )

        signals.m2m_changed.send(
            sender=self.through,
            action="post_add",
            instance=self.instance,
            reverse=False,
            model=self.through.tag_model(),
            pk_set=new_ids,
            using=db,
        )

    def _to_tag_model_instances(self, tags, tag_kwargs):
        """
        Takes an iterable containing either strings, tag objects, or a mixture
        of both and returns set of tag objects.
        """
        db = router.db_for_write(self.through, instance=self.instance)

        str_tags = set()
        tag_objs = set()

        for t in tags:
            if isinstance(t, self.through.tag_model()):
                tag_objs.add(t)
            elif isinstance(t, str):
                str_tags.add(t)
            else:
                raise ValueError(
                    "Cannot add {} ({}). Expected {} or str.".format(
                        t, type(t), type(self.through.tag_model())
                    )
                )

        case_insensitive = getattr(settings, "TAGGIT_CASE_INSENSITIVE", False)
        manager = self.through.tag_model()._default_manager.using(db)

        if case_insensitive:
            # Some databases can do case-insensitive comparison with IN, which
            # would be faster, but we can't rely on it or easily detect it.
            existing = []
            tags_to_create = []

            for name in str_tags:
                try:
                    tag = manager.get(name__iexact=name)
                    existing.append(tag)
                except self.through.tag_model().DoesNotExist:
                    tags_to_create.append(name)
        else:
            # If str_tags has 0 elements Django actually optimizes that to not
            # do a query.  Malcolm is very smart.
            existing = manager.filter(name__in=str_tags, **tag_kwargs)

            tags_to_create = str_tags - set(t.name for t in existing)

        tag_objs.update(existing)

        for new_tag in tags_to_create:
            if case_insensitive:
                lookup = {"name__iexact": new_tag, **tag_kwargs}
            else:
                lookup = {"name": new_tag, **tag_kwargs}

            tag, create = manager.get_or_create(**lookup, defaults={"name": new_tag})
            tag_objs.add(tag)

        return tag_objs

    @require_instance_manager
    def names(self):
        return self.get_queryset().values_list("name", flat=True)

    @require_instance_manager
    def slugs(self):
        return self.get_queryset().values_list("slug", flat=True)

    @require_instance_manager
    def set(self, *tags, **kwargs):
        """
        Set the object's tags to the given n tags. If the clear kwarg is True
        then all existing tags are removed (using `.clear()`) and the new tags
        added. Otherwise, only those tags that are not present in the args are
        removed and any new tags added.

        Any kwarg apart from 'clear' will be passed when adding tags.

        """
        db = router.db_for_write(self.through, instance=self.instance)

        clear = kwargs.pop("clear", False)
        tag_kwargs = kwargs.pop("tag_kwargs", {})

        if clear:
            self.clear()
            self.add(*tags, **kwargs)
        else:
            # make sure we're working with a collection of a uniform type
            objs = self._to_tag_model_instances(tags, tag_kwargs)

            # get the existing tag strings
            old_tag_strs = set(
                self.through._default_manager.using(db)
                .filter(**self._lookup_kwargs())
                .values_list("tag__name", flat=True)
            )

            new_objs = []
            for obj in objs:
                if obj.name in old_tag_strs:
                    old_tag_strs.remove(obj.name)
                else:
                    new_objs.append(obj)

            self.remove(*old_tag_strs)
            self.add(*new_objs, **kwargs)

    @require_instance_manager
    def remove(self, *tags):
        if not tags:
            return

        db = router.db_for_write(self.through, instance=self.instance)

        qs = (
            self.through._default_manager.using(db)
            .filter(**self._lookup_kwargs())
            .filter(tag__name__in=tags)
        )

        old_ids = set(qs.values_list("tag_id", flat=True))

        signals.m2m_changed.send(
            sender=self.through,
            action="pre_remove",
            instance=self.instance,
            reverse=False,
            model=self.through.tag_model(),
            pk_set=old_ids,
            using=db,
        )
        qs.delete()
        signals.m2m_changed.send(
            sender=self.through,
            action="post_remove",
            instance=self.instance,
            reverse=False,
            model=self.through.tag_model(),
            pk_set=old_ids,
            using=db,
        )

    @require_instance_manager
    def clear(self):
        db = router.db_for_write(self.through, instance=self.instance)

        signals.m2m_changed.send(
            sender=self.through,
            action="pre_clear",
            instance=self.instance,
            reverse=False,
            model=self.through.tag_model(),
            pk_set=None,
            using=db,
        )

        self.through._default_manager.using(db).filter(**self._lookup_kwargs()).delete()

        signals.m2m_changed.send(
            sender=self.through,
            action="post_clear",
            instance=self.instance,
            reverse=False,
            model=self.through.tag_model(),
            pk_set=None,
            using=db,
        )

    def most_common(self, min_count=None, extra_filters=None):
        queryset = (
            self.get_queryset(extra_filters)
            .annotate(num_times=models.Count(self.through.tag_relname()))
            .order_by("-num_times")
        )
        if min_count:
            queryset = queryset.filter(num_times__gte=min_count)

        return queryset

    @require_instance_manager
    def similar_objects(self):
        lookup_kwargs = self._lookup_kwargs()
        lookup_keys = sorted(lookup_kwargs)
        qs = self.through.objects.values(*lookup_kwargs.keys())
        qs = qs.annotate(n=models.Count("pk"))
        qs = qs.exclude(**lookup_kwargs)
        qs = qs.filter(tag__in=self.all())
        qs = qs.order_by("-n")

        # TODO: This all feels like a bit of a hack.
        items = {}
        if len(lookup_keys) == 1:
            # Can we do this without a second query by using a select_related()
            # somehow?
            f = self.through._meta.get_field(lookup_keys[0])
            remote_field = f.remote_field
            rel_model = remote_field.model
            objs = rel_model._default_manager.filter(
                **{
                    "%s__in"
                    % remote_field.field_name: [r["content_object"] for r in qs]
                }
            )
            for obj in objs:
                items[(getattr(obj, remote_field.field_name),)] = obj
        else:
            preload = {}
            for result in qs:
                preload.setdefault(result["content_type"], set())
                preload[result["content_type"]].add(result["object_id"])

            for ct, obj_ids in preload.items():
                ct = ContentType.objects.get_for_id(ct)
                for obj in ct.model_class()._default_manager.filter(pk__in=obj_ids):
                    items[(ct.pk, obj.pk)] = obj

        results = []
        for result in qs:
            obj = items[tuple(result[k] for k in lookup_keys)]
            obj.similar_tags = result["n"]
            results.append(obj)
        return results


class TaggableManager(RelatedField):
    # Field flags
    many_to_many = True
    many_to_one = False
    one_to_many = False
    one_to_one = False

    _related_name_counter = 0

    def __init__(
        self,
        verbose_name=_("Tags"),
        help_text=_("A comma-separated list of tags."),
        through=None,
        blank=False,
        related_name=None,
        to=None,
        manager=_TaggableManager,
    ):
        self.through = through or TaggedItem

        rel = ManyToManyRel(self, to, related_name=related_name, through=self.through)

        super().__init__(
            verbose_name=verbose_name,
            help_text=help_text,
            blank=blank,
            null=True,
            serialize=False,
            rel=rel,
        )

        self.swappable = False
        self.manager = manager

    def __get__(self, instance, model):
        if instance is not None and instance.pk is None:
            raise ValueError(
                "%s objects need to have a primary key value "
                "before you can access their tags." % model.__name__
            )
        return self.manager(
            through=self.through,
            model=model,
            instance=instance,
            prefetch_cache_name=self.name,
        )

    def deconstruct(self):
        """
        Deconstruct the object, used with migrations.
        """
        name, path, args, kwargs = super().deconstruct()
        # Remove forced kwargs.
        for kwarg in ("serialize", "null"):
            del kwargs[kwarg]
        # Add arguments related to relations.
        # Ref: https://github.com/jazzband/django-taggit/issues/206#issuecomment-37578676
        rel = self.remote_field
        if isinstance(rel.through, str):
            kwargs["through"] = rel.through
        elif not rel.through._meta.auto_created:
            kwargs["through"] = "{}.{}".format(
                rel.through._meta.app_label, rel.through._meta.object_name
            )

        related_model = rel.model
        if isinstance(related_model, str):
            kwargs["to"] = related_model
        else:
            kwargs["to"] = "{}.{}".format(
                related_model._meta.app_label, related_model._meta.object_name
            )

        return name, path, args, kwargs

    def contribute_to_class(self, cls, name):
        self.set_attributes_from_name(name)
        self.model = cls
        self.opts = cls._meta

        cls._meta.add_field(self)
        setattr(cls, name, self)
        if not cls._meta.abstract:
            if isinstance(self.remote_field.model, str):

                def resolve_related_class(cls, model, field):
                    field.remote_field.model = model

                lazy_related_operation(
                    resolve_related_class, cls, self.remote_field.model, field=self
                )
            if isinstance(self.through, str):

                def resolve_related_class(cls, model, field):
                    self.through = model
                    self.remote_field.through = model
                    self.post_through_setup(cls)

                lazy_related_operation(
                    resolve_related_class, cls, self.through, field=self
                )
            else:
                self.post_through_setup(cls)

    def get_internal_type(self):
        return "ManyToManyField"

    def post_through_setup(self, cls):
        self.use_gfk = self.through is None or issubclass(
            self.through, CommonGenericTaggedItemBase
        )

        if not self.remote_field.model:
            self.remote_field.model = self.through._meta.get_field(
                "tag"
            ).remote_field.model

        if self.use_gfk:
            tagged_items = GenericRelation(self.through)
            tagged_items.contribute_to_class(cls, "tagged_items")

        for rel in cls._meta.local_many_to_many:
            if rel == self or not isinstance(rel, TaggableManager):
                continue
            if rel.through == self.through:
                raise ValueError(
                    "You can't have two TaggableManagers with the"
                    " same through model."
                )

    def save_form_data(self, instance, value):
        getattr(instance, self.name).set(*value)

    def formfield(self, form_class=TagField, **kwargs):
        defaults = {
            "label": capfirst(self.verbose_name),
            "help_text": self.help_text,
            "required": not self.blank,
        }
        defaults.update(kwargs)
        return form_class(**defaults)

    def value_from_object(self, obj):
        if obj.pk is None:
            return []
        qs = self.through.objects.select_related("tag").filter(
            **self.through.lookup_kwargs(obj)
        )
        return [ti.tag for ti in qs]

    def m2m_reverse_name(self):
        return self.through._meta.get_field("tag").column

    def m2m_reverse_field_name(self):
        return self.through._meta.get_field("tag").name

    def m2m_target_field_name(self):
        return self.model._meta.pk.name

    def m2m_reverse_target_field_name(self):
        return self.remote_field.model._meta.pk.name

    def m2m_column_name(self):
        if self.use_gfk:
            return self.through._meta.private_fields[0].fk_field
        return self.through._meta.get_field("content_object").column

    def m2m_db_table(self):
        return self.through._meta.db_table

    def bulk_related_objects(self, new_objs, using):
        return []

    def _get_mm_case_path_info(self, direct=False, filtered_relation=None):
        pathinfos = []
        linkfield1 = self.through._meta.get_field("content_object")
        linkfield2 = self.through._meta.get_field(self.m2m_reverse_field_name())
        if direct:
            if VERSION < (2, 0):
                join1infos = linkfield1.get_reverse_path_info()
                join2infos = linkfield2.get_path_info()
            else:
                join1infos = linkfield1.get_reverse_path_info(
                    filtered_relation=filtered_relation
                )
                join2infos = linkfield2.get_path_info(
                    filtered_relation=filtered_relation
                )
        else:
            if VERSION < (2, 0):
                join1infos = linkfield2.get_reverse_path_info()
                join2infos = linkfield1.get_path_info()
            else:
                join1infos = linkfield2.get_reverse_path_info(
                    filtered_relation=filtered_relation
                )
                join2infos = linkfield1.get_path_info(
                    filtered_relation=filtered_relation
                )
        pathinfos.extend(join1infos)
        pathinfos.extend(join2infos)
        return pathinfos

    def _get_gfk_case_path_info(self, direct=False, filtered_relation=None):
        pathinfos = []
        from_field = self.model._meta.pk
        opts = self.through._meta
        linkfield = self.through._meta.get_field(self.m2m_reverse_field_name())
        if direct:
            if VERSION < (2, 0):
                join1infos = [
                    PathInfo(
                        self.model._meta,
                        opts,
                        [from_field],
                        self.remote_field,
                        True,
                        False,
                    )
                ]
                join2infos = linkfield.get_path_info()
            else:
                join1infos = [
                    PathInfo(
                        self.model._meta,
                        opts,
                        [from_field],
                        self.remote_field,
                        True,
                        False,
                        filtered_relation,
                    )
                ]
                join2infos = linkfield.get_path_info(
                    filtered_relation=filtered_relation
                )
        else:
            if VERSION < (2, 0):
                join1infos = linkfield.get_reverse_path_info()
                join2infos = [
                    PathInfo(opts, self.model._meta, [from_field], self, True, False)
                ]
            else:
                join1infos = linkfield.get_reverse_path_info(
                    filtered_relation=filtered_relation
                )
                join2infos = [
                    PathInfo(
                        opts,
                        self.model._meta,
                        [from_field],
                        self,
                        True,
                        False,
                        filtered_relation,
                    )
                ]
        pathinfos.extend(join1infos)
        pathinfos.extend(join2infos)
        return pathinfos

    def get_path_info(self, filtered_relation=None):
        if self.use_gfk:
            return self._get_gfk_case_path_info(
                direct=True, filtered_relation=filtered_relation
            )
        else:
            return self._get_mm_case_path_info(
                direct=True, filtered_relation=filtered_relation
            )

    def get_reverse_path_info(self, filtered_relation=None):
        if self.use_gfk:
            return self._get_gfk_case_path_info(
                direct=False, filtered_relation=filtered_relation
            )
        else:
            return self._get_mm_case_path_info(
                direct=False, filtered_relation=filtered_relation
            )

    def get_joining_columns(self, reverse_join=False):
        if reverse_join:
            return ((self.model._meta.pk.column, "object_id"),)
        else:
            return (("object_id", self.model._meta.pk.column),)

    def get_extra_restriction(self, where_class, alias, related_alias):
        extra_col = self.through._meta.get_field("content_type").column
        content_type_ids = [
            ContentType.objects.get_for_model(subclass).pk
            for subclass in _get_subclasses(self.model)
        ]
        return ExtraJoinRestriction(related_alias, extra_col, content_type_ids)

    def get_reverse_joining_columns(self):
        return self.get_joining_columns(reverse_join=True)

    @property
    def related_fields(self):
        return [(self.through._meta.get_field("object_id"), self.model._meta.pk)]

    @property
    def foreign_related_fields(self):
        return [self.related_fields[0][1]]


def _get_subclasses(model):
    subclasses = [model]
    for field in model._meta.get_fields():
        if isinstance(field, OneToOneRel) and getattr(
            field.field.remote_field, "parent_link", None
        ):
            subclasses.extend(_get_subclasses(field.related_model))
    return subclasses



from django.db import models

from scarlet.versioning.models import VersionModel, Cloneable, BaseModel


class NameModel(models.Model):
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True
        app_label = "version_twomodels"


class AuthorBase(BaseModel):
    class Meta:
        app_label = "version_twomodels"


class Author(VersionModel, NameModel):
    associates = models.ManyToManyField(AuthorBase, blank=True)

    class Meta:
        app_label = "version_twomodels"
        _base_model = AuthorBase

    def __unicode__(self):
        return self.name


class BookBase(BaseModel):
    class Meta:
        app_label = "version_twomodels"


class Book(VersionModel, NameModel):
    _clone_related = ["review", "galleries"]

    author = models.ForeignKey(AuthorBase, on_delete=models.CASCADE)
    galleries = models.ManyToManyField("Gallery")

    class Meta:
        app_label = "version_twomodels"
        _base_model = BookBase


class Review(Cloneable):
    book = models.ForeignKey(Book, on_delete=models.CASCADE)
    text = models.CharField(max_length=255)

    class Meta:
        app_label = "version_twomodels"


class StoreBase(BaseModel):
    class Meta:
        app_label = "version_twomodels"


class Store(VersionModel, NameModel):
    books = models.ManyToManyField(BookBase)

    class Meta:
        _base_model = StoreBase
        app_label = "version_twomodels"


class Gallery(Cloneable):
    name = models.CharField(max_length=255)

    class Meta:
        app_label = "version_twomodels"



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import IntegrityError, models, router, transaction
from django.utils.text import slugify
from django.utils.translation import gettext, gettext_lazy as _

try:
    from unidecode import unidecode
except ImportError:

    def unidecode(tag):
        return tag


class TagBase(models.Model):
    name = models.CharField(verbose_name=_("name"), unique=True, max_length=100)
    slug = models.SlugField(verbose_name=_("slug"), unique=True, max_length=100)

    def __str__(self):
        return self.name

    def __gt__(self, other):
        return self.name.lower() > other.name.lower()

    def __lt__(self, other):
        return self.name.lower() < other.name.lower()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if self._state.adding and not self.slug:
            self.slug = self.slugify(self.name)
            using = kwargs.get("using") or router.db_for_write(
                type(self), instance=self
            )
            # Make sure we write to the same db for all attempted writes,
            # with a multi-master setup, theoretically we could try to
            # write and rollback on different DBs
            kwargs["using"] = using
            # Be opportunistic and try to save the tag, this should work for
            # most cases ;)
            try:
                with transaction.atomic(using=using):
                    res = super().save(*args, **kwargs)
                return res
            except IntegrityError:
                pass
            # Now try to find existing slugs with similar names
            slugs = set(
                type(self)
                ._default_manager.filter(slug__startswith=self.slug)
                .values_list("slug", flat=True)
            )
            i = 1
            while True:
                slug = self.slugify(self.name, i)
                if slug not in slugs:
                    self.slug = slug
                    # We purposely ignore concurrency issues here for now.
                    # (That is, till we found a nice solution...)
                    return super().save(*args, **kwargs)
                i += 1
        else:
            return super().save(*args, **kwargs)

    def slugify(self, tag, i=None):
        slug = slugify(unidecode(tag))
        if i is not None:
            slug += "_%d" % i
        return slug


class Tag(TagBase):
    class Meta:
        verbose_name = _("tag")
        verbose_name_plural = _("tags")
        app_label = "taggit"


class ItemBase(models.Model):
    def __str__(self):
        return gettext("%(object)s tagged with %(tag)s") % {
            "object": self.content_object,
            "tag": self.tag,
        }

    class Meta:
        abstract = True

    @classmethod
    def tag_model(cls):
        field = cls._meta.get_field("tag")
        return field.remote_field.model

    @classmethod
    def tag_relname(cls):
        field = cls._meta.get_field("tag")
        return field.remote_field.related_name

    @classmethod
    def lookup_kwargs(cls, instance):
        return {"content_object": instance}

    @classmethod
    def tags_for(cls, model, instance=None, **extra_filters):
        kwargs = extra_filters or {}
        if instance is not None:
            kwargs.update({"%s__content_object" % cls.tag_relname(): instance})
            return cls.tag_model().objects.filter(**kwargs)
        kwargs.update({"%s__content_object__isnull" % cls.tag_relname(): False})
        return cls.tag_model().objects.filter(**kwargs).distinct()


class TaggedItemBase(ItemBase):
    tag = models.ForeignKey(
        Tag, related_name="%(app_label)s_%(class)s_items", on_delete=models.CASCADE
    )

    class Meta:
        abstract = True


class CommonGenericTaggedItemBase(ItemBase):
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        verbose_name=_("content type"),
        related_name="%(app_label)s_%(class)s_tagged_items",
    )
    content_object = GenericForeignKey()

    class Meta:
        abstract = True

    @classmethod
    def lookup_kwargs(cls, instance):
        return {
            "object_id": instance.pk,
            "content_type": ContentType.objects.get_for_model(instance),
        }

    @classmethod
    def tags_for(cls, model, instance=None, **extra_filters):
        tag_relname = cls.tag_relname()
        model = model._meta.concrete_model
        kwargs = {
            "%s__content_type__app_label" % tag_relname: model._meta.app_label,
            "%s__content_type__model" % tag_relname: model._meta.model_name,
        }
        if instance is not None:
            kwargs["%s__object_id" % tag_relname] = instance.pk
        if extra_filters:
            kwargs.update(extra_filters)
        return cls.tag_model().objects.filter(**kwargs).distinct()


class GenericTaggedItemBase(CommonGenericTaggedItemBase):
    object_id = models.IntegerField(verbose_name=_("object ID"), db_index=True)

    class Meta:
        abstract = True


class GenericUUIDTaggedItemBase(CommonGenericTaggedItemBase):
    object_id = models.UUIDField(verbose_name=_("object ID"), db_index=True)

    class Meta:
        abstract = True


class TaggedItem(GenericTaggedItemBase, TaggedItemBase):
    class Meta:
        verbose_name = _("tagged item")
        verbose_name_plural = _("tagged items")
        app_label = "taggit"
        index_together = [["content_type", "object_id"]]
        unique_together = [["content_type", "object_id", "tag"]]



from .. import Provider as PhoneNumberProvider


class Provider(PhoneNumberProvider):
    phonenumber_prefixes = [134, 135, 136, 137, 138, 139, 147, 150,
                            151, 152, 157, 158, 159, 182, 187, 188,
                            130, 131, 132, 145, 155, 156, 185, 186,
                            145, 133, 153, 180, 181, 189]
    formats = [str(i) + "########" for i in phonenumber_prefixes]

    def phonenumber_prefix(self):
        return self.random_element(self.phonenumber_prefixes)



from django.db.backends.base.base import NO_DB_ALIAS
from django.db.backends.postgresql.base import (
    DatabaseWrapper as Psycopg2DatabaseWrapper,
)

from .features import DatabaseFeatures
from .introspection import PostGISIntrospection
from .operations import PostGISOperations
from .schema import PostGISSchemaEditor


class DatabaseWrapper(Psycopg2DatabaseWrapper):
    SchemaEditorClass = PostGISSchemaEditor

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if kwargs.get('alias', '') != NO_DB_ALIAS:
            self.features = DatabaseFeatures(self)
            self.ops = PostGISOperations(self)
            self.introspection = PostGISIntrospection(self)

    def prepare_database(self):
        super().prepare_database()
        # Check that postgis extension is installed.
        with self.cursor() as cursor:
            cursor.execute("CREATE EXTENSION IF NOT EXISTS postgis")



try:
    from django.conf.urls import url, patterns
except ImportError:
    from django.conf.urls.defaults import url, patterns

import views

urlpatterns = patterns("", url(r"^$", views.fancy_view, name="fancy_view"),)



import requests
from functools import partial
from typing import Optional, List, Tuple

from django.conf import settings

BASE_URL = "https://api.cloudflare.com/client/v4"
AUTH_HEADERS = {
    "X-Auth-Email": settings.CLOUDFLARE["AUTH_EMAIL"],
    "X-Auth-Key": settings.CLOUDFLARE["AUTH_KEY"],
}


def purge_remote_cache(urls: Optional[List[str]] = None) -> Tuple[bool, str]:
    """
    Function to purge saved pages on remote cache.
    At the moment only Cloudflare is supported.
    """
    url = f"{BASE_URL}/zones/{settings.CLOUDFLARE['ZONE_ID']}/purge_cache"
    send_request = partial(requests.post, url=url, headers=AUTH_HEADERS, timeout=6.0)

    if urls:
        r = send_request(data={"files": urls})
    else:
        r = send_request(data={"purge_everything": True})

    if r.ok and r.data.get("success"):
        return True, "Cache flushed successfully"
    elif r.ok:
        return False, f"Something went wrong. Errors: {' '.join(r.data.get('errors'))}"
    else:
        return False, f"Something went wrong. Status code: {r.status_code}"



# code snippet copied from https://gist.github.com/NotSqrt/5f3c76cd15e40ef62d09
from pytest_django.lazy_django import get_django_version


class DisableMigrations(object):
    def __init__(self):
        self._django_version = get_django_version()

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        if self._django_version >= (1, 9):
            return None
        else:
            return "notmigrations"



from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS, connections, migrations
from django.db.migrations.loader import AmbiguityError, MigrationLoader
from django.db.migrations.migration import SwappableTuple
from django.db.migrations.optimizer import MigrationOptimizer
from django.db.migrations.writer import MigrationWriter
from django.utils.version import get_docs_version


class Command(BaseCommand):
    help = "Squashes an existing set of migrations (from first until specified) into a single new one."

    def add_arguments(self, parser):
        parser.add_argument(
            'app_label',
            help='App label of the application to squash migrations for.',
        )
        parser.add_argument(
            'start_migration_name', nargs='?',
            help='Migrations will be squashed starting from and including this migration.',
        )
        parser.add_argument(
            'migration_name',
            help='Migrations will be squashed until and including this migration.',
        )
        parser.add_argument(
            '--no-optimize', action='store_true',
            help='Do not try to optimize the squashed operations.',
        )
        parser.add_argument(
            '--noinput', '--no-input', action='store_false', dest='interactive',
            help='Tells Django to NOT prompt the user for input of any kind.',
        )
        parser.add_argument(
            '--squashed-name',
            help='Sets the name of the new squashed migration.',
        )
        parser.add_argument(
            '--no-header', action='store_false', dest='include_header',
            help='Do not add a header comment to the new squashed migration.',
        )

    def handle(self, **options):

        self.verbosity = options['verbosity']
        self.interactive = options['interactive']
        app_label = options['app_label']
        start_migration_name = options['start_migration_name']
        migration_name = options['migration_name']
        no_optimize = options['no_optimize']
        squashed_name = options['squashed_name']
        include_header = options['include_header']
        # Validate app_label.
        try:
            apps.get_app_config(app_label)
        except LookupError as err:
            raise CommandError(str(err))
        # Load the current graph state, check the app and migration they asked for exists
        loader = MigrationLoader(connections[DEFAULT_DB_ALIAS])
        if app_label not in loader.migrated_apps:
            raise CommandError(
                "App '%s' does not have migrations (so squashmigrations on "
                "it makes no sense)" % app_label
            )

        migration = self.find_migration(loader, app_label, migration_name)

        # Work out the list of predecessor migrations
        migrations_to_squash = [
            loader.get_migration(al, mn)
            for al, mn in loader.graph.forwards_plan((migration.app_label, migration.name))
            if al == migration.app_label
        ]

        if start_migration_name:
            start_migration = self.find_migration(loader, app_label, start_migration_name)
            start = loader.get_migration(start_migration.app_label, start_migration.name)
            try:
                start_index = migrations_to_squash.index(start)
                migrations_to_squash = migrations_to_squash[start_index:]
            except ValueError:
                raise CommandError(
                    "The migration '%s' cannot be found. Maybe it comes after "
                    "the migration '%s'?\n"
                    "Have a look at:\n"
                    "  python manage.py showmigrations %s\n"
                    "to debug this issue." % (start_migration, migration, app_label)
                )

        # Tell them what we're doing and optionally ask if we should proceed
        if self.verbosity > 0 or self.interactive:
            self.stdout.write(self.style.MIGRATE_HEADING("Will squash the following migrations:"))
            for migration in migrations_to_squash:
                self.stdout.write(" - %s" % migration.name)

            if self.interactive:
                answer = None
                while not answer or answer not in "yn":
                    answer = input("Do you wish to proceed? [yN] ")
                    if not answer:
                        answer = "n"
                        break
                    else:
                        answer = answer[0].lower()
                if answer != "y":
                    return

        # Load the operations from all those migrations and concat together,
        # along with collecting external dependencies and detecting
        # double-squashing
        operations = []
        dependencies = set()
        # We need to take all dependencies from the first migration in the list
        # as it may be 0002 depending on 0001
        first_migration = True
        for smigration in migrations_to_squash:
            if smigration.replaces:
                raise CommandError(
                    "You cannot squash squashed migrations! Please transition "
                    "it to a normal migration first: "
                    "https://docs.djangoproject.com/en/%s/topics/migrations/#squashing-migrations" % get_docs_version()
                )
            operations.extend(smigration.operations)
            for dependency in smigration.dependencies:
                if isinstance(dependency, SwappableTuple):
                    if settings.AUTH_USER_MODEL == dependency.setting:
                        dependencies.add(("__setting__", "AUTH_USER_MODEL"))
                    else:
                        dependencies.add(dependency)
                elif dependency[0] != smigration.app_label or first_migration:
                    dependencies.add(dependency)
            first_migration = False

        if no_optimize:
            if self.verbosity > 0:
                self.stdout.write(self.style.MIGRATE_HEADING("(Skipping optimization.)"))
            new_operations = operations
        else:
            if self.verbosity > 0:
                self.stdout.write(self.style.MIGRATE_HEADING("Optimizing..."))

            optimizer = MigrationOptimizer()
            new_operations = optimizer.optimize(operations, migration.app_label)

            if self.verbosity > 0:
                if len(new_operations) == len(operations):
                    self.stdout.write("  No optimizations possible.")
                else:
                    self.stdout.write(
                        "  Optimized from %s operations to %s operations." %
                        (len(operations), len(new_operations))
                    )

        # Work out the value of replaces (any squashed ones we're re-squashing)
        # need to feed their replaces into ours
        replaces = []
        for migration in migrations_to_squash:
            if migration.replaces:
                replaces.extend(migration.replaces)
            else:
                replaces.append((migration.app_label, migration.name))

        # Make a new migration with those operations
        subclass = type("Migration", (migrations.Migration,), {
            "dependencies": dependencies,
            "operations": new_operations,
            "replaces": replaces,
        })
        if start_migration_name:
            if squashed_name:
                # Use the name from --squashed-name.
                prefix, _ = start_migration.name.split('_', 1)
                name = '%s_%s' % (prefix, squashed_name)
            else:
                # Generate a name.
                name = '%s_squashed_%s' % (start_migration.name, migration.name)
            new_migration = subclass(name, app_label)
        else:
            name = '0001_%s' % (squashed_name or 'squashed_%s' % migration.name)
            new_migration = subclass(name, app_label)
            new_migration.initial = True

        # Write out the new migration file
        writer = MigrationWriter(new_migration, include_header)
        with open(writer.path, "w", encoding='utf-8') as fh:
            fh.write(writer.as_string())

        if self.verbosity > 0:
            self.stdout.write(self.style.MIGRATE_HEADING("Created new squashed migration %s" % writer.path))
            self.stdout.write("  You should commit this migration but leave the old ones in place;")
            self.stdout.write("  the new migration will be used for new installs. Once you are sure")
            self.stdout.write("  all instances of the codebase have applied the migrations you squashed,")
            self.stdout.write("  you can delete them.")
            if writer.needs_manual_porting:
                self.stdout.write(self.style.MIGRATE_HEADING("Manual porting required"))
                self.stdout.write("  Your migrations contained functions that must be manually copied over,")
                self.stdout.write("  as we could not safely copy their implementation.")
                self.stdout.write("  See the comment at the top of the squashed migration for details.")

    def find_migration(self, loader, app_label, name):
        try:
            return loader.get_migration_by_prefix(app_label, name)
        except AmbiguityError:
            raise CommandError(
                "More than one migration matches '%s' in app '%s'. Please be "
                "more specific." % (name, app_label)
            )
        except KeyError:
            raise CommandError(
                "Cannot find a migration matching '%s' from app '%s'." %
                (name, app_label)
            )



import datetime
import re
import sys
from decimal import Decimal

from toml.decoder import InlineTableDict

if sys.version_info >= (3,):
    unicode = str


def dump(o, f, encoder=None):
    """Writes out dict as toml to a file

    Args:
        o: Object to dump into toml
        f: File descriptor where the toml should be stored
        encoder: The ``TomlEncoder`` to use for constructing the output string

    Returns:
        String containing the toml corresponding to dictionary

    Raises:
        TypeError: When anything other than file descriptor is passed
    """

    if not f.write:
        raise TypeError("You can only dump an object to a file descriptor")
    d = dumps(o, encoder=encoder)
    f.write(d)
    return d


def dumps(o, encoder=None):
    """Stringifies input dict as toml

    Args:
        o: Object to dump into toml
        encoder: The ``TomlEncoder`` to use for constructing the output string

    Returns:
        String containing the toml corresponding to dict

    Examples:
        ```python
        >>> import toml
        >>> output = {
        ... 'a': "I'm a string",
        ... 'b': ["I'm", "a", "list"],
        ... 'c': 2400
        ... }
        >>> toml.dumps(output)
        'a = "I\'m a string"\nb = [ "I\'m", "a", "list",]\nc = 2400\n'
        ```
    """

    retval = ""
    if encoder is None:
        encoder = TomlEncoder(o.__class__)
    addtoretval, sections = encoder.dump_sections(o, "")
    retval += addtoretval
    outer_objs = [id(o)]
    while sections:
        section_ids = [id(section) for section in sections.values()]
        for outer_obj in outer_objs:
            if outer_obj in section_ids:
                raise ValueError("Circular reference detected")
        outer_objs += section_ids
        newsections = encoder.get_empty_table()
        for section in sections:
            addtoretval, addtosections = encoder.dump_sections(
                sections[section], section)

            if addtoretval or (not addtoretval and not addtosections):
                if retval and retval[-2:] != "\n\n":
                    retval += "\n"
                retval += "[" + section + "]\n"
                if addtoretval:
                    retval += addtoretval
            for s in addtosections:
                newsections[section + "." + s] = addtosections[s]
        sections = newsections
    return retval


def _dump_str(v):
    if sys.version_info < (3,) and hasattr(v, 'decode') and isinstance(v, str):
        v = v.decode('utf-8')
    v = "%r" % v
    if v[0] == 'u':
        v = v[1:]
    singlequote = v.startswith("'")
    if singlequote or v.startswith('"'):
        v = v[1:-1]
    if singlequote:
        v = v.replace("\\'", "'")
        v = v.replace('"', '\\"')
    v = v.split("\\x")
    while len(v) > 1:
        i = -1
        if not v[0]:
            v = v[1:]
        v[0] = v[0].replace("\\\\", "\\")
        # No, I don't know why != works and == breaks
        joinx = v[0][i] != "\\"
        while v[0][:i] and v[0][i] == "\\":
            joinx = not joinx
            i -= 1
        if joinx:
            joiner = "x"
        else:
            joiner = "u00"
        v = [v[0] + joiner + v[1]] + v[2:]
    return unicode('"' + v[0] + '"')


def _dump_float(v):
    return "{}".format(v).replace("e+0", "e+").replace("e-0", "e-")


def _dump_time(v):
    utcoffset = v.utcoffset()
    if utcoffset is None:
        return v.isoformat()
    # The TOML norm specifies that it's local time thus we drop the offset
    return v.isoformat()[:-6]


class TomlEncoder(object):

    def __init__(self, _dict=dict, preserve=False):
        self._dict = _dict
        self.preserve = preserve
        self.dump_funcs = {
            str: _dump_str,
            unicode: _dump_str,
            list: self.dump_list,
            bool: lambda v: unicode(v).lower(),
            int: lambda v: v,
            float: _dump_float,
            Decimal: _dump_float,
            datetime.datetime: lambda v: v.isoformat().replace('+00:00', 'Z'),
            datetime.time: _dump_time,
            datetime.date: lambda v: v.isoformat()
        }

    def get_empty_table(self):
        return self._dict()

    def dump_list(self, v):
        retval = "["
        for u in v:
            retval += " " + unicode(self.dump_value(u)) + ","
        retval += "]"
        return retval

    def dump_inline_table(self, section):
        """Preserve inline table in its compact syntax instead of expanding
        into subsection.

        https://github.com/toml-lang/toml#user-content-inline-table
        """
        retval = ""
        if isinstance(section, dict):
            val_list = []
            for k, v in section.items():
                val = self.dump_inline_table(v)
                val_list.append(k + " = " + val)
            retval += "{ " + ", ".join(val_list) + " }\n"
            return retval
        else:
            return unicode(self.dump_value(section))

    def dump_value(self, v):
        # Lookup function corresponding to v's type
        dump_fn = self.dump_funcs.get(type(v))
        if dump_fn is None and hasattr(v, '__iter__'):
            dump_fn = self.dump_funcs[list]
        # Evaluate function (if it exists) else return v
        return dump_fn(v) if dump_fn is not None else self.dump_funcs[str](v)

    def dump_sections(self, o, sup):
        retstr = ""
        if sup != "" and sup[-1] != ".":
            sup += '.'
        retdict = self._dict()
        arraystr = ""
        for section in o:
            section = unicode(section)
            qsection = section
            if not re.match(r'^[A-Za-z0-9_-]+$', section):
                qsection = _dump_str(section)
            if not isinstance(o[section], dict):
                arrayoftables = False
                if isinstance(o[section], list):
                    for a in o[section]:
                        if isinstance(a, dict):
                            arrayoftables = True
                if arrayoftables:
                    for a in o[section]:
                        arraytabstr = "\n"
                        arraystr += "[[" + sup + qsection + "]]\n"
                        s, d = self.dump_sections(a, sup + qsection)
                        if s:
                            if s[0] == "[":
                                arraytabstr += s
                            else:
                                arraystr += s
                        while d:
                            newd = self._dict()
                            for dsec in d:
                                s1, d1 = self.dump_sections(d[dsec], sup +
                                                            qsection + "." +
                                                            dsec)
                                if s1:
                                    arraytabstr += ("[" + sup + qsection +
                                                    "." + dsec + "]\n")
                                    arraytabstr += s1
                                for s1 in d1:
                                    newd[dsec + "." + s1] = d1[s1]
                            d = newd
                        arraystr += arraytabstr
                else:
                    if o[section] is not None:
                        retstr += (qsection + " = " +
                                   unicode(self.dump_value(o[section])) + '\n')
            elif self.preserve and isinstance(o[section], InlineTableDict):
                retstr += (qsection + " = " +
                           self.dump_inline_table(o[section]))
            else:
                retdict[qsection] = o[section]
        retstr += arraystr
        return (retstr, retdict)


class TomlPreserveInlineDictEncoder(TomlEncoder):

    def __init__(self, _dict=dict):
        super(TomlPreserveInlineDictEncoder, self).__init__(_dict, True)


class TomlArraySeparatorEncoder(TomlEncoder):

    def __init__(self, _dict=dict, preserve=False, separator=","):
        super(TomlArraySeparatorEncoder, self).__init__(_dict, preserve)
        if separator.strip() == "":
            separator = "," + separator
        elif separator.strip(' \t\n\r,'):
            raise ValueError("Invalid separator for arrays")
        self.separator = separator

    def dump_list(self, v):
        t = []
        retval = "["
        for u in v:
            t.append(self.dump_value(u))
        while t != []:
            s = []
            for u in t:
                if isinstance(u, list):
                    for r in u:
                        s.append(r)
                else:
                    retval += " " + unicode(u) + self.separator
            t = s
        retval += "]"
        return retval


class TomlNumpyEncoder(TomlEncoder):

    def __init__(self, _dict=dict, preserve=False):
        import numpy as np
        super(TomlNumpyEncoder, self).__init__(_dict, preserve)
        self.dump_funcs[np.float16] = _dump_float
        self.dump_funcs[np.float32] = _dump_float
        self.dump_funcs[np.float64] = _dump_float
        self.dump_funcs[np.int16] = self._dump_int
        self.dump_funcs[np.int32] = self._dump_int
        self.dump_funcs[np.int64] = self._dump_int

    def _dump_int(self, v):
        return "{}".format(int(v))


class TomlPreserveCommentEncoder(TomlEncoder):

    def __init__(self, _dict=dict, preserve=False):
        from toml.decoder import CommentValue
        super(TomlPreserveCommentEncoder, self).__init__(_dict, preserve)
        self.dump_funcs[CommentValue] = lambda v: v.dump(self.dump_value)


class TomlPathlibEncoder(TomlEncoder):

    def _dump_pathlib_path(self, v):
        return _dump_str(str(v))

    def dump_value(self, v):
        if (3, 4) <= sys.version_info:
            import pathlib
            if isinstance(v, pathlib.PurePath):
                v = str(v)
        return super(TomlPathlibEncoder, self).dump_value(v)



import taggit.managers
from django.db import migrations, models

import scarlet.assets.fields
import scarlet.assets.utils


class Migration(migrations.Migration):

    dependencies = [
        ("taggit", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Asset",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                (
                    "file",
                    scarlet.assets.fields.AssetRealFileField(
                        upload_to=scarlet.assets.utils.assets_dir
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        db_index=True,
                        max_length=255,
                        choices=[
                            (b"unknown", b"Unknown"),
                            (b"image", b"Image"),
                            (b"document", b"Document"),
                            (b"audio", b"Audio"),
                            (b"video", b"Video"),
                        ],
                    ),
                ),
                ("slug", models.SlugField(unique=True, max_length=255)),
                ("user_filename", models.CharField(max_length=255)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("modified", models.DateTimeField(auto_now=True)),
                ("cbversion", models.PositiveIntegerField(editable=False)),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        to="taggit.Tag",
                        through="taggit.TaggedItem",
                        help_text="A comma-separated list of tags.",
                        verbose_name="Tags",
                    ),
                ),
            ],
            options={"abstract": False,},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ImageDetail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("width", models.PositiveIntegerField()),
                ("height", models.PositiveIntegerField()),
                ("name", models.CharField(max_length=255)),
                ("editable", models.BooleanField(default=False, editable=False)),
                ("x", models.PositiveIntegerField(null=True)),
                ("x2", models.PositiveIntegerField(null=True)),
                ("y", models.PositiveIntegerField(null=True)),
                ("y2", models.PositiveIntegerField(null=True)),
                (
                    "image",
                    models.ForeignKey(
                        to="assets.Asset", on_delete=models.deletion.CASCADE,
                    ),
                ),
            ],
            options={"abstract": False,},
            bases=(models.Model,),
        ),
    ]



from .. import Provider as PersonProvider


class Provider(PersonProvider):
    formats_female = (
        '{{first_name_female}} {{last_name}}',
        '{{first_name_female}} {{last_name}}',
        '{{first_name_female}} {{last_name}}',
        '{{first_name_female}} {{last_name}}',
        '{{first_name_female}} {{last_name}}',
        '{{prefix_female}} {{first_name_female}} {{last_name}}',
    )

    formats_male = (
        '{{first_name_male}} {{last_name}}',
        '{{first_name_male}} {{last_name}}',
        '{{first_name_male}} {{last_name}}',
        '{{first_name_male}} {{last_name}}',
        '{{first_name_male}} {{last_name}}',
        '{{prefix_male}} {{first_name_male}} {{last_name}}',

    )

    formats = formats_male + formats_female

    """
    To a previous (undocumented?) list of female given names was added the 100
    most popular names in Brazil in 2014 and 2015 according to Exame magazine:
    * http://exame.abril.com.br/brasil/noticias/os-100-nomes-mais-comuns-no-brasil-em-2014
    * http://exame.abril.com.br/brasil/noticias/os-100-nomes-mais-comuns-no-brasil-em-2015
    """
    first_names_female = (
        'Agatha', 'Alana', 'Alexia', 'Alice', 'Alícia', 'Amanda',
        'Ana Beatriz', 'Ana Carolina', 'Ana Clara', 'Ana Julia', 'Ana Júlia',
        'Ana Laura', 'Ana Luiza', 'Ana Lívia', 'Ana Sophia', 'Ana Vitória',
        'Ana', 'Beatriz', 'Bianca', 'Brenda', 'Bruna', 'Bárbara', 'Camila',
        'Carolina', 'Caroline', 'Catarina', 'Cecília', 'Clara', 'Clarice',
        'Daniela', 'Eduarda', 'Elisa', 'Eloah', 'Emanuella', 'Emanuelly',
        'Emilly', 'Esther', 'Evelyn', 'Fernanda', 'Gabriela', 'Gabrielly',
        'Giovanna', 'Helena', 'Heloísa', 'Isabel', 'Isabella', 'Isabelly',
        'Isadora', 'Isis', 'Joana', 'Julia', 'Juliana', 'Júlia', 'Kamilly',
        'Lara', 'Larissa', 'Laura', 'Lavínia', 'Laís', 'Letícia', 'Lorena',
        'Luana', 'Luiza', 'Luna', 'Lívia', 'Maitê', 'Manuela', 'Marcela',
        'Maria Alice', 'Maria Cecília', 'Maria Clara', 'Maria Eduarda',
        'Maria Fernanda', 'Maria Julia', 'Maria Luiza', 'Maria Sophia',
        'Maria Vitória', 'Maria', 'Mariana', 'Mariane', 'Marina', 'Maysa',
        'Melissa', 'Milena', 'Mirella', 'Natália', 'Nicole', 'Nina', 'Olivia',
        'Pietra', 'Rafaela', 'Raquel', 'Rebeca', 'Sabrina', 'Sarah', 'Sofia',
        'Sophia', 'Sophie', 'Stella', 'Stephany', 'Valentina', 'Vitória',
        'Yasmin',
    )

    """
    To a previous (undocumented?) list of male given names was added the 100
    most popular names in Brazil in 2014 and 2015 according to this blog post:
    * http://exame.abril.com.br/brasil/noticias/os-100-nomes-mais-comuns-no-brasil-em-2014
    * http://exame.abril.com.br/brasil/noticias/os-100-nomes-mais-comuns-no-brasil-em-2015
    """
    first_names_male = (
        'Alexandre', 'André', 'Anthony', 'Antônio', 'Arthur', 'Augusto',
        'Benjamin', 'Benício', 'Bernardo', 'Breno', 'Bruno', 'Bryan', 'Caio',
        'Calebe', 'Carlos Eduardo', 'Cauã', 'Cauê', 'Daniel', 'Danilo',
        'Davi Lucas', 'Davi Lucca', 'Davi Luiz', 'Davi', 'Diego', 'Diogo',
        'Eduardo', 'Emanuel', 'Enrico', 'Enzo Gabriel', 'Enzo', 'Erick',
        'Felipe', 'Fernando', 'Francisco', 'Gabriel', 'Guilherme',
        'Gustavo Henrique', 'Gustavo', 'Heitor', 'Henrique', 'Ian', 'Igor',
        'Isaac', 'Joaquim', 'João Felipe', 'João Gabriel', 'João Guilherme',
        'João Lucas', 'João Miguel', 'João Pedro', 'João Vitor', 'João',
        'Juan', 'Kaique', 'Kevin', 'Leandro', 'Leonardo', 'Levi', 'Lorenzo',
        'Lucas Gabriel', 'Lucas', 'Lucca', 'Luigi', 'Luiz Felipe',
        'Luiz Fernando', 'Luiz Gustavo', 'Luiz Henrique', 'Luiz Miguel',
        'Luiz Otávio', 'Marcelo', 'Marcos Vinicius', 'Matheus', 'Miguel',
        'Murilo', 'Nathan', 'Nicolas', 'Noah', 'Otávio', 'Paulo',
        'Pedro Henrique', 'Pedro Lucas', 'Pedro Miguel', 'Pedro', 'Pietro',
        'Rafael', 'Raul', 'Renan', 'Rodrigo', 'Ryan', 'Samuel', 'Thales',
        'Theo', 'Thiago', 'Thomas', 'Vicente', 'Vinicius', 'Vitor Gabriel',
        'Vitor Hugo', 'Vitor', 'Yago', 'Yuri',
    )

    first_names = first_names_male + first_names_female

    """
    To a previous (undocumented?) list of family names was added the 70
    most popular family names in Brazil according to this blog post:
    * http://nomeschiques.com/os-70-sobrenomes-mais-comuns-e-famosos-do-brasil/
    """
    last_names = (
        'Almeida', 'Alves', 'Aragão', 'Araújo', 'Azevedo', 'Barbosa', 'Barros',
        'Caldeira', 'Campos', 'Cardoso', 'Cardoso', 'Carvalho', 'Castro',
        'Cavalcanti', 'Correia', 'Costa', 'Costela', 'Cunha', 'da Conceição',
        'da Costa', 'da Cruz', 'da Cunha', 'da Luz', 'da Mata', 'da Mota',
        'da Paz', 'da Rocha', 'da Rosa', 'das Neves', 'Dias', 'Duarte',
        'Farias', 'Fernandes', 'Ferreira', 'Fogaça', 'Freitas', 'Gomes',
        'Gonçalves', 'Jesus', 'Lima', 'Lopes', 'Martins', 'Melo', 'Mendes',
        'Monteiro', 'Moraes', 'Moreira', 'Moura', 'Nascimento', 'Nogueira',
        'Novaes', 'Nunes', 'Oliveira', 'Peixoto', 'Pereira', 'Pinto', 'Pires',
        'Porto', 'Ramos', 'Rezende', 'Ribeiro', 'Rocha', 'Rodrigues', 'Sales',
        'Santos', 'Silva', 'Silveira', 'Souza', 'Teixeira', 'Viana', 'Vieira',
    )

    prefixes_female = ('Srta.', 'Sra.', 'Dra.')
    prefixes_male = ('Sr.', 'Dr.')



from ..ar_AA import Provider as ArabicDateTimeProvider


class Provider(ArabicDateTimeProvider):
    MONTH_NAMES = {
        '01': 'يناير',
        '02': 'فبراير',
        '03': 'مارس',
        '04': 'أبريل',
        '05': 'مايو',
        '06': 'يونيو',
        '07': 'يوليو',
        '08': 'أغسطس',
        '09': 'سبتمبر',
        '10': 'أكتوبر',
        '11': 'نوفمبر',
        '12': 'ديسمبر',
    }



from collections import OrderedDict

from ..en_PH import Provider as EnPhProvider


class Provider(EnPhProvider):
    """
    Provider for company names for fil_PH locale

    Companies in the Philippines rarely have Filipino names, and when they do, the English name is usually used way more
    frequently by the locals. In some cases, the Filipino names are more like in Taglish, so for the purposes of this
    provider, only English company names will be generated for this locale.

    Company and brand taglines in pure Filipino, however, are much more common, so this provider will generate catch
    phrases in pure Filipino randomly alongside the English ones.
    """

    catch_phrase_formats = OrderedDict([
        ('{{english_catch_phrase}}', 0.64),
        ('Ang {{random_noun_ish_good_trait}} ng {{random_object_of_concern}}!', 0.12),
        ('Serbisyong {{random_good_service_adjective}} para sa {{random_object_of_concern}}!', 0.12),
        ('Kahit kailan, {{random_good_service_adjective_chain}}!', 0.12),
    ])
    noun_ish_good_traits = (
        'bida',
        'ginhawa',
        'haligi',
        'karangalan',
        'lingkod',
        'liwanag',
        'numero uno',
        'pag-asa',
        'tulay',
    )
    good_service_adjectives = (
        'bida',
        'dekalidad',
        'hindi umaatras',
        'kakaiba',
        'maasahan',
        'magaling',
        'mapatitiwalaan',
        'numero uno',
        'panalo',
        'tagumpay',
        'tama',
        'tapat',
        'totoo',
        'tunay',
        'walang kapantay',
        'walang katulad',
        'walang tatalo',
    )
    objects_of_concern = [
        'Filipino',
        'Pilipinas',
        'Pilipino',
        'Pinoy',
        'bahay',
        'bansa',
        'bayan',
        'buhay',
        'mamamayan',
        'mundo',
        'tahanan',
    ]

    def random_noun_ish_good_trait(self):
        return self.random_element(self.noun_ish_good_traits)

    def random_good_service_adjective(self):
        return self.random_element(self.good_service_adjectives)

    def random_good_service_adjective_chain(self):
        adjectives = self.random_elements(self.good_service_adjectives, length=2, unique=True)
        return ' at '.join(adjectives)

    def random_object_of_concern(self):
        return self.random_element(self.objects_of_concern)

    def english_catch_phrase(self):
        return super().catch_phrase()

    def catch_phrase(self):
        return self.random_element(self.catch_phrase_formats)
